#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import functools
import multiprocessing as mp
import os
import re

import dill
import numpy as np
from scipy import integrate as ig
from scipy import interpolate as ip
from scipy import stats
from scipy.optimize import newton
from .constants import post15140
from .constants import pre15140


def get_mesa_version(mesa_dir):
    """
    Get the version of MESA in mesa_dir

    Args:
        mesa_dir (str): Path to MESA directory.

    Returns:
        str: Version string

    Examples:
        >>> get_mesa_version(os.environ['MESA_DIR'])
        '24.03.1'
    """
    with open(f'{mesa_dir}/data/version_number', 'r') as handle:
        version = handle.read().strip()
    return version


def compare_version(version1, version2, operator):
    """
    Compare two MESA versions.

    Args:
        version1 (str): LHS of comparison.
        version2 (str): RHS of comparison.
        operator (str): Comparison operator string. Can be one of '<', '>', '<=', '>=', '==', '!='.

    Returns:
        bool: Result of performing the comparison.

    Examples:
        >>> compare_version('11701', '8118', '<')  # '11701' < '8118' would return True.
        False
    """
    allowed_ops = ['<', '>', '<=', '>=', '==', '!=']
    if not operator in allowed_ops:
        raise ValueError(f'`operator` must be one of: {", ".join(allowed_ops)}.')
    i = allowed_ops.index(operator)

    r_version1 = False
    r_version2 = False
    if version1.startswith('r'):
        r_version1 = True
    if version2.startswith('r'):
        r_version2 = True

    lt = None
    gt = None
    lte = None
    gte = None
    eq = None
    neq = None

    eq = version1 == version2
    neq = not eq

    # if version number starts with an r it is after 15140
    if r_version1 and r_version2:
        lt = version1 < version2
    elif r_version1 and not r_version2:
        lt = False
    elif not r_version1 and r_version2:
        lt = True
    else:
        lt = int(version1) < int(version2)

    gte = not lt
    gt = gte and neq
    lte = not gt

    return [lt, gt, lte, gte, eq, neq][i]


def get_constants(p_or_hist):
    """
    Get the correct MESA constants used for p_or_hist.

    Args:
        p_or_hist (Profile or History): Profile or History for which to get constants

    Returns:
        module: Module containing the constants for p_or_hist.
    """
    version = str(p_or_hist.header['version_number'])

    if compare_version(version, '15140', '<'):
        return pre15140
    else:
        return post15140


# Mixing type codes for pre and post 15140
mix_dict = {'pre15140': {-1:'no_region',
                         0: 'no_mixing',
                         1: 'convective_mixing',
                         2: 'softened_convective_mixing',
                         3: 'overshoot_mixing',
                         4: 'semiconvective_mixing',
                         5: 'thermohaline_mixing',
                         6: 'rotation_mixing',
                         7: 'rayleigh_taylor_mixing',
                         8: 'minimum_mixing',
                         9: 'anonymous_mixing'},
            'post15140': {-1:'no_region',
                          0: 'no_mixing',
                          1: 'convective_mixing',
                          2: 'overshoot_mixing',
                          3: 'semiconvective_mixing',
                          4: 'thermohaline_mixing',
                          5: 'rotation_mixing',
                          6: 'rayleigh_taylor_mixing',
                          7: 'minimum_mixing',
                          8: 'anonymous_mixing',
                          9: 'leftover_convective_mixing'},
            # Need a merged when combining histories from different versions
            'merged': {'no_region': -1,
                       'no_mixing': 100,
                       'convective_mixing': 101,
                       'overshoot_mixing': 103,
                       'semiconvective_mixing': 104,
                       'thermohaline_mixing': 105,
                       'rotation_mixing': 106,
                       'minimum_mixing': 107,
                       'anonymous_mixing': 109},
            'names':  {'no_region': '',
                       'no_mixing': 'No mixing',
                       'convective_mixing': 'Convection',
                       'overshoot_mixing': 'Overshooting',
                       'semiconvective_mixing': 'Semiconvection',
                       'thermohaline_mixing': 'Thermohaline',
                       'rotation_mixing': 'Rotational',
                       'minimum_mixing': 'Minimum',
                       'anonymous_mixing': 'Anonymous'},
            }
mix_dict['merged_r'] = {v:k for k,v in mix_dict['merged'].items()}  # Reversed merged.
def convert_mixing_type(mix_type, version, unknown_mixing=100, other_mixing=[(None, None)]):
    """
    Convert the mixing type codes to a merged version compatible with pre- and post-15140 MESA.

    Args:
        mix_type (np.array):
        version (str): MESA version
        unknown_mixing (int, optional): Mixing type code to use for unreckognized mixing type code.
            Defaults to 100, which is the merged code for no_mixing.
        other_mixing (list of tuple(str, int), optional):  Information on custom mixing types. (name, mix_type_code)

    Returns:
        np.array: New mixing type codes associated with mix_type.
    """
    if compare_version(version, '15140', '>='):
        pre_post = 'post'
    else:
        pre_post = 'pre'
    key = f'{pre_post}15140'

    merged = mix_dict['merged']
    for custom_name, custom_mix_type in other_mixing:
        if custom_name is not None and custom_mix_type is not None:
            mix_dict[key][custom_mix_type] = custom_name
            merged[custom_name] = custom_mix_type
            mix_dict['merged_r'] = {v: k for k, v in mix_dict['merged'].items()}

    mix_names = np.vectorize(mix_dict[key].get)(mix_type, unknown_mixing)
    return np.vectorize(mix_dict['merged'].get)(mix_names, unknown_mixing)


def cell2face(val, dm, dm_is_m=False, m_center=0):
    """
    Convert a cell average value (e.g. density) to the value at the cell face (e.g. radius).

    Args:
        val (np.array): Value to convert.
        dm (np.array): Cell mass or mass coordinate if dm_is_m is True.
        dm_is_m (bool, optional): If True, treat dm as mass coordinate instead of cell mass.
        m_center (float, optional: If dm_is_m is True, set this as the interior mass.

    Returns:
        np.array: Value at cell face of val.
    """
    if dm_is_m:
        dm = np.diff(dm, append=m_center)
    face = np.zeros_like(val)
    face[0] = val[0]
    face[1:] = (val[:-1] * dm[1:] + val[1:] * dm[:-1]) / (dm[1:] + dm[:-1])
    return face


def get_logTeffL(hist, mask=None, linear=False):
    if 'log_Teff' in hist.columns:
        logTeff = hist.get('log_Teff')
    elif 'effective_T' in hist.columns:
        logTeff = np.log10(hist.get('effective_T'))
    else:
        raise ValueError('log_Teff or effective_T not in history file.')

    if 'photosphere_L' in hist.columns:
        logL = np.log10(hist.get('photosphere_L'))
    elif 'luminosity' in hist.columns:
        logL = np.log10(hist.get('luminosity'))
    elif 'log_L' in hist.columns:
        logL = hist.get('log_L')
    else:
        if 'photosphere_r' in hist.columns:
            print('Calculating logL from Teff and R_photosphere.')
            R = hist.get('photosphere_r')
            logL = 2 * np.log10(R) + 4 * logTeff - 4 * np.log10(5777)
        else:
            raise ValueError('log_L, luminosity, or photosphere_L not in history file.')
    mask = get_mask(hist, mask)
    if linear:
        logTeff = 10 ** logTeff
        logL = 10 ** logL
    return logTeff[mask], logL[mask]


def get_radius(p, unit='cm'):
    """
    Get the radius from Profile p and perform a unit conversion with the correct constants.

    Args:
        p (Profile):
        unit (str):

    Returns:
        array-like: Radial coordinate in unit.
    """
    c = get_constants(p)

    if 'radius' in p.columns:
        radius = p.get('radius') * c.rsun
    elif 'logR' in p.columns:
        radius = 10 ** p.get('logR') * c.rsun
    elif 'radius_cm' in p.columns:
        radius = p.get('radius_cm')
    else:
        raise ValueError("No radius found in profile.")

    if unit == 'log':
        radius = np.log10(radius/c.rsun)
    elif unit.lower().replace('_', '') in ['rsun', 'rsol']:
        radius = radius / c.rsun
    elif unit.lower() == 'cm':
        radius = radius
    else:
        raise ValueError(f'Unknown unit {unit}. Must be one of cm, rsun, rsol, or log.')
    return radius


def get_m_bot_CZ(hist, mask=None, max_q_bot=0.999):
    """
    Get the mass coordinate of the bottom of the convective envelope.

    Args:
        hist (History):
        mask (bool, np.array, or function, optional): If True, will exclude pre-main sequence.
                If an array of bools will use that as mask. If a function, will call function(hist) and use that as the mask.
        max_q_bot (float, optional): Maximum dimensionless mass coordinate below which to look for the bottom of the convective envelope.

    Returns:
        array-like: Bottom of the convective envelope.
    """
    if 'm_botCZ' in hist.columns:
        return hist.get('m_botCZ', mask=mask)
    m_bot_CZ = np.zeros_like(hist.get('star_mass', mask=mask))
    n_mix = 0
    for col in hist.columns:
        if col.startswith('mix_qtop'):
            n_mix += 1
    for i in range(n_mix):
        i += 1
        mix_type = convert_mixing_type(hist.get(f'mix_type_{i}', mask=mask), hist.header['version_number'])
        m_bot_CZ = np.maximum(m_bot_CZ,
                              hist.get('star_mass', mask=mask) * hist.get(f'mix_qtop_{i}', mask=mask) *
                              (mix_type == mix_dict['merged']['no_mixing']) * (hist.get(f'mix_qtop_{i}', mask=mask) <= max_q_bot))
    return m_bot_CZ


def get_cheb_mask(hist, min_Yc=1e-6, max_Yc=1.0):
    """
    Get the Core Helium Burning mask for hist.

    Defined as:
    * Have helium in core
    * Have convective core
    * Hydrogen depleted core

    Args:
        hist (History):
        min_Yc (float, optional): Minimum central helium mass fraction. Defaults to 1e-6.
        max_Yc (float, optional): Maximum central helium mass fraction. Defaults to 1.0.

    Returns:
        array of bool:
    """

    mask = hist.data.center_he4 >= min_Yc
    mask = np.logical_and(mask, hist.data.center_he4 <= max_Yc)
    mask = np.logical_and(mask, hist.data.mass_conv_core > 0)
    mask = np.logical_and(mask, hist.data.center_h1 < 1e-6)

    return mask


def get_rc_mask(hist, min_Yc=0.1, max_fYc=0.95, first_chunk=False):
    """
    Get the Red Clump mask for hist.

    Defined as:
    Have helium in core
    Have convective core
    Hydrogen depleted core

    Args:
        hist (History):
        min_Yc (float, optional): Minimum central helium mass fraction. Defaults to 0.1.
        max_fYc (float, optional): Maximum central helium mass fraction factor. This factor is multiplied by the
            maximum Yc achieved during evolution to get the value which is compared against. Defaults to 0.95.
        first_chunk (bool, optional): If True, only returns the mask for the first contiguous block of True in the mask.

    Returns:
        array of bool:
    """

    mask = hist.data.center_he4 >= min_Yc
    mask = np.logical_and(mask, hist.data.center_he4 <= max_fYc * max(hist.data.center_he4))
    mask = np.logical_and(mask, hist.data.mass_conv_core > 0)
    mask = np.logical_and(mask, hist.data.center_h1 < 1e-6)

    if first_chunk and np.any(mask):
        end_first_block = np.where(np.diff(mask, prepend=False))[0][1]
        mask[end_first_block + 1:] = False
    return mask


def get_pms_mask(hist, invert=False, ZAMS_method='Xc', fXc=0.99):
    """
    Get the pre-main sequence mask for hist.

    Args:
        hist (History):
        invert (bool, optional): If True, invert the PMS mask. Defaults to False.
        ZAMS_method (str or callable, optional): Can be 'Xc', or 'LH'. If 'Xc', will use when X_c reaches ``fXc`` of
            the initial central hydrogen mass fraction. If 'LH', will end when the hydrogen burning luminosity is equal
            to total luminosity. If ZAMS_method is callable will call ``ZAMS_method(hist)``, which should return a
            mask with the same length as hist. Defaults to 'Xc'.
        fXc (float, optional): Define the end of PMS as when central hydrogen mass fraction reaches ``fXc`` of the
            initial value. Defaults to 0.99

    Returns:
        array of bool:

    """

    if ZAMS_method == 'LH':
        _, logL = get_logTeffL(hist)
        if 'log_LH' in hist.columns:
            maskL = ((hist.data.log_LH - logL) < 0) & (hist.data.center_h1 > 0.6)
        else:
            have_pp = 'pp' in hist.columns
            have_cno = 'cno' in hist.columns
            if have_pp and have_cno:
                log_LH = np.log10(10**hist.get('pp') + 10**hist.get('cno'))
                maskL = ((log_LH - logL) < 0) & (hist.data.center_h1 > 0.6)
            else:
                raise KeyError('Must have either log_LH or pp and cno as columns if using hydrogen luminosity as ZAMS start.')
        starts_pms = maskL[0]

        mask = np.zeros_like(maskL)
        if starts_pms:
            end = np.where(np.diff(maskL))[0][0]
            mask[:end + 1] = True
    elif ZAMS_method == 'Xc':
        mask = hist.data.center_h1 >= hist.data.center_h1[0] * 0.99
    elif callable(ZAMS_method):
        mask = ZAMS_method(hist)
    else:
        raise ValueError(f'Unknown ZAMS_method: {ZAMS_method}.')


    if invert:
        mask = ~mask

    return mask


def get_ms_mask(hist, min_Xc=1e-3, ZAMS_method='Xc'):
    """
    Get the main sequence mask for hist.

    Args:
        hist (History):
        min_Xc (float): Minimum central hydrogen mass fraction for the end of the MS.
        ZAMS_method (str, optional): See get_pms_mask for details. Defaults to 'Xc'.

    Returns:
        array of bool:

    """

    mask = hist.data.center_h1 > min_Xc
    mask = np.logical_and(mask, get_pms_mask(hist, invert=True, ZAMS_method=ZAMS_method))

    return mask


def get_sgb_mask(hist, min_Xc=1e-3, fCZ=0.35):
    """
    Get the sub-giant branch mask.

    Args:
        hist (History):
        min_Xc (float): Minimum central hydrogen mass fraction for the end of the MS.
        fCZ (float): Convective envelope mass fraction limit for end of SGB.

    Returns:
        array of bool:

    """
    ms_mask = get_ms_mask(hist, min_Xc)
    if np.any(ms_mask):
        min_mod = hist.data.model_number[ms_mask][-1]
    else:
        if 'center_Rho' in hist.columns:
            center_Rho = hist.get('center_Rho')
        elif 'log_center_Rho' in hist.columns:
            center_Rho = 10 ** hist.get('log_center_Rho')
        if np.log10(center_Rho[0]) >= 3.5:  # starts during or after SGB
            return np.zeros_like(ms_mask, dtype=bool)
        min_mod = hist.data.model_number[0]

    mask = get_m_bot_CZ(hist) / hist.data.star_mass <= (1 - fCZ)
    mask = mask & (hist.data.model_number > min_mod)
    max_mod = hist.data.model_number[mask][0]

    mask = np.logical_and(hist.data.model_number > min_mod, hist.data.model_number < max_mod)
    return mask


def get_bump_mask(hist, max_logT_lim=3.8, min_logT_lim=3.6, min_logL=0.5):
    """
    Get the RGB bump mask.

    Defined as having an increasing effective temperature on the RGB within the Teff and L limits.

    Args:
        hist (History):
        max_logT_lim (float, optional): Maximum logT below which to look for the RGBb. Defaults to 3.8.
        min_logT_lim (float, optional): Minimum logT below which to look for the RGBb. Defaults to 3.6.
        min_logL (float, optional): Minimum logL  above which to look for the RGBb. Defaults to 0.5.

    Returns:
        array of bool:

    """

    logTeff, logL = get_logTeffL(hist)

    mask = hist.data.center_h1 < 1e-9
    mask = mask & (hist.data.center_he4 > 0.95)
    mask = mask & get_rgb_mask(hist)

    dlogT = np.diff(logTeff, prepend=99)
    dlogL = np.diff(logL, prepend=99)

    angle = np.arctan2(dlogL, dlogT)
    mask = mask & (angle < -0.2 * np.pi) & (angle > -0.6 * np.pi)

    # Cooler than (3.5, 1) -> (3.6, 3)
    y = 20 * (logTeff - 3.5) + 1
    mask = mask & (y > logL)

    # Bump is defined as when L decreases and Teff increases on RGB
    # submask_bump = np.logical_and(np.diff(logL[mask], prepend=-99) < 0, np.diff(logTeff[mask], prepend=99) > 0)
    # Bump is defined as when Teff increases on RGB
    # submask_bump = np.diff(logTeff[mask], prepend=99) > 0
    # mask[mask] = submask_bump

    return mask


def get_tip_mask(hist, logT_lim=3.8, min_logL=2.0):
    """
    Get the RGB tip mask for hist.

    Args:
        hist (History):
        logT_lim (float, optional): Maximum logT below which to look for the RGB tip.

    Returns:
        array of bool:

    """
    logTeff, logL = get_logTeffL(hist)

    mask = hist.data.center_h1 < 1e-9
    mask = np.logical_and(mask, hist.data.center_he4 > 0.95)
    mask = np.logical_and(mask, logTeff < logT_lim)
    mask = np.logical_and(mask, logL > min_logL)

    if sum(mask) == 0:
        return np.zeros_like(mask, dtype=bool)

    minT = min(logTeff[mask])
    maxL = max(logL[mask])

    mask = np.logical_and(mask, logTeff < minT + 0.01)
    mask = np.logical_and(mask, logL > maxL - 0.1)

    return mask


def get_rgb_mask(hist, min_Xc=1e-3, logT_lim=3.8):
    """
    Get the RGB mask for hist.

    Args:
        hist (History):
        min_Xc (float, optional): Minimum central hydrogen mass fraction for the end of the MS, passed to get_sgb_mask.
            Defaults to 1e-3.
        logT_lim (float, optional): Maximum logT below which to look for the end of the RGB. Defaults to 3.8.

    Returns:
        array of bool:

    """
    try:
        sgb_mask = get_sgb_mask(hist, min_Xc)
        if np.any(sgb_mask):
            first_mod = max(hist.data.model_number[sgb_mask])
        else:
            first_mod = hist.data.model_number[0]
    except IndexError:
        first_mod = hist.data.model_number[0]

    has_CHeB = np.any(get_cheb_mask(hist))

    if has_CHeB:
        tip_mask = get_tip_mask(hist, logT_lim)
        logTeff, logL = get_logTeffL(hist)
        masked_i_min_T = np.argmin(logTeff[tip_mask])
        last_mod = hist.data.model_number[tip_mask][masked_i_min_T]
    else:
        if hist.data.center_he4[-1] > 0.9:  # stopped on RGB
            last_mod = hist.data.model_number[-1]

    mask = np.logical_and(hist.data.model_number > first_mod, hist.data.model_number <= last_mod)
    return mask


def get_flashes_mask(hist, logT_lim=3.8):
    """
    Get the helium flashes mask.
    Starts after the RGB tip and ends when CHeB starts.

    Args:
        hist (History):
        logT_lim (float, optional): Passed to `get_tip_mask` to get the start of He-flashes.

    Returns:
        array of bool:

    """
    rc_mask = get_cheb_mask(hist)
    if np.any(rc_mask):
        last_mod = min(hist.data.model_number[rc_mask])
    else:
        last_mod = hist.data.model_number[-1]

    tip_mask = get_tip_mask(hist, logT_lim)
    logTeff, logL = get_logTeffL(hist)
    if np.any(tip_mask):
        masked_i_min_T = np.argmin(logTeff[tip_mask])
        first_mod = hist.data.model_number[tip_mask][masked_i_min_T]
        mask = np.logical_and(hist.data.model_number > first_mod, hist.data.model_number < last_mod)
    else:
        mask = np.zeros_like(rc_mask)

    return mask


# def get_agb_mask(hist):
#     """
#     Get the AGB mask.
#
#
#     Args:
#         hist (History):
#
#     Returns:
#         array of bool:
#     TODO: Define an ending condition.
#     """
#     rc_mask = get_cheb_mask(hist)
#     if np.any(rc_mask):
#         mod_agb_start = max(hist.data.model_number[rc_mask])
#         mask = hist.data.model_number > mod_agb_start
#     else:
#         mask = np.zeros_like(rc_mask, dtype=bool)
#     return mask


mask_functions = [get_pms_mask, get_ms_mask, get_sgb_mask, get_rgb_mask, get_bump_mask, get_tip_mask, get_flashes_mask,
                  get_cheb_mask, get_rc_mask]#, get_agb_mask]
mask_names = ['PMS', 'MS', 'SGB', 'RGB', 'RGBb', 'RGB tip', 'He flashes', 'CHeB', 'RC']#, 'post-CHeB']


def get_gridnum(hist):
    """

    Args:
        hist (History):

    Returns:

    """
    return int(os.path.split(hist.LOGS)[0][-4:])


def get_mask(hist, use_mask):
    """
    Evaluate a mask.

    Args:
        hist (History):
        use_mask (bool, np.array, or function, optional): If True, will exclude pre-main sequence.
                If an array of bools will use that as mask. If a function, will call function(hist) and use that as the mask.

    Returns:
        array of bool:

    """
    if hasattr(use_mask, '__len__'):
        if len(use_mask) == len(hist.get('model_number')):
            mask = use_mask
        else:
            # Try to get numpy to raise the index error first
            _ = hist.get('model_number')[use_mask]
            raise IndexError(
                f'Length of mask not the same as length of data ({len(use_mask)} vs {len(hist.get("model_number"))}')
    elif callable(use_mask):
        mask = use_mask(hist)
    else:
        if use_mask:
            mask = get_pms_mask(hist, invert=True)
        else:
            mask = np.ones_like(hist.get('model_number'), dtype=bool)
    return mask


def get_mean(hist, name, use_mask=None, domain='star_age', filter=None, get_std=False, verbose=True):
    """
    Calculate the average of a quantity defined by `name` over `domain`.

    Args:
        hist (History):
        name (str, array, function): Quantity in hist to average. Can be a column name in hist, an array with the same
                length as hist, an array with the mask specified by use_mask already applied, or a function with
                signature function(hist, mask).
        use_mask (bool, np.array, or function, optional): If True, will exclude pre-main sequence.
                If an array of bools will use that as mask. If a function, will call function(hist) and use that as the mask.
        domain (str, array, optional): Domain over which to average. Must be monotonically increasing or
                decreasing. Can be a column name in hist, or an array with the same length as hist. Defaults to 'star_age'.
        filter (str, optional): Quantities to mask out. Can be one of 'negative', '-', '-ve', 'positive', '+', '+ve', 'finite', 'inf', 'nan', None. Defaults to None.
        get_std (bool, optional): If True, will also return the standard deviation.

    Returns:
        float or Tuple of floats: The average
    """
    if use_mask is not None:
        mask = get_mask(hist, use_mask=use_mask)
    else:
        mask = ...

    if isinstance(domain, str):
        xdat = hist.get(domain)[mask]
    else:
        xdat = domain[mask]
    x_monotonic = False
    if np.all(np.diff(xdat) > 0):
        x_monotonic = True
    elif np.all(np.diff(xdat) < 0):
        x_monotonic = True
    else:
        raise ValueError(f"domain `{domain}` must be monotonic.")
    if isinstance(name, str):
        ydat = hist.get(name)[mask]
    elif callable(name):
        ydat = name(hist, mask)
    else:
        if not isinstance(mask, Ellipsis.__class__):
            if len(name) == sum(mask):  # mask already applied
                ydat = name
            else:
                ydat = name[mask]
        else:
            ydat = name

    if filter is not None:
        filters = filter.split()
        mask = np.ones_like(ydat, dtype=bool)
        for filter in filters:
            if filter.lower() in ['negative', '-', '-ve']:
                mask = mask & (ydat >= 0)
            elif filter.lower() in ['positive', '+', '+ve']:
                mask = mask & (ydat <= 0)
            elif filter.lower() in ['finite', 'inf', 'nan']:
                mask = mask & (np.isfinite(ydat))

        xdat = xdat[mask]
        ydat = ydat[mask]

    if len(xdat) == 0:
        if verbose:
            print(f"Warning: 0-length data when calculating mean of {domain}")
        mean = np.nan
        std = np.nan
    else:
        mean = ig.trapz(ydat, xdat) / (xdat[-1] - xdat[0])
        std = (ig.trapz((ydat - mean) ** 2, xdat) / (xdat[-1] - xdat[0])) ** 0.5
        std = np.mean(np.diff(
            np.quantile(ip.interp1d(xdat, ydat)(np.linspace(min(xdat), max(xdat), 201)), [0.15865, 0.50, 0.84135])))
    if get_std:
        return mean, std
    else:
        return mean


def get_weighted_quantile(x, w, q=(0.15865, 0.50, 0.84135)):
    """
    Calculate the weighted quantile of `x` using weights `w`.

    Args:
        x:
        w (array): Weights of `x`.
        q (tuple of floats): Quantiles at which to calculate.

    Returns:

    """
    if not np.all(x.shape == w.shape):
        raise ValueError('x and w must have the same shape.')

    if len(x) == 0:
        return np.nan * np.ones_like(q)

    argsort = np.argsort(x)
    x = x[argsort]
    w = w[argsort]
    cumsumw = np.cumsum(w)
    cumsumw /= cumsumw[-1]

    return ip.interp1d(cumsumw, x)(q)


# def get_instability_strip(Z, logL_min=3, logL_max=4.5):
#     """Bono2005"""
#     Z = min(max(Z, 0.004), 0.02)
#
#     blue_coefs = np.array([[0.004, 3.617, 0.108, -0.018, 0.001],
#                           [.008, 3.557, 0.153, -0.026, 0.001],
#                           [.02, 3.729, 0.064, -0.017, 0.001]])
#     red_coefs = np.array([[0.004, 3.678, 0.090, -0.026, 0.001],
#                          [0.008, 3.709, 0.077, -0.022, 0.001],
#                          [0.02, 3.763, 0.053, -0.021, 0.001]])
#     logL = np.linspace(logL_min, logL_max, 101)
#     edges = []
#     for coefs in [blue_coefs, red_coefs]:
#         alpha, beta, gamma, _ = ip.interp1d(coefs[:, 0], coefs[:, 1:].T)(Z)
#         logT = alpha + beta*logL + gamma*logL**2
#         edges.append(logT)
#     return logL, edges

def get_instability_strip(Z, Y, logL_min=3, logL_max=4.5, kind='RRLyrae'):
    if kind == 'RRLyrae':
        # https://ui.adsabs.harvard.edu/abs/2015ApJ...808...50M/abstract
        logL_min = 1.5
        logL_max = 1.9

        logL = np.linspace(logL_min, logL_max, 101)
        logTr = -0.084 * logL - 0.012 * np.log10(Z) + 3.879
        logTb = -0.080 * logL - 0.012 * np.log10(Z) + 3.957

        return logL, (logTb, logTr)

    elif kind == 'Cepheid':
        # https://arxiv.org/pdf/astro-ph/9801242.pdf

        (Xref, Zref) = (0.7, 0.004)
        Yref = 1 - Xref - Zref
        deltaY = Y - Yref
        deltaZ = Z - Zref

        logL = np.linspace(logL_min, logL_max, 101)
        logTeff = -0.036 * logL + 3.925
        dlogTeff = 0.04 * deltaY - 0.49 * deltaZ
        logTeff += dlogTeff

    return logL, (logTeff, logTeff - 0.06)


def get_evo_phase(hist, phase_funcs):
    """

    Args:
        hist:
        phase_funcs:

    Returns:

    """
    evo_phase = np.zeros_like(hist.get('model_number'))
    max_phase = 1

    # TODO: Be able to deal with runs that end early
    phase_pass = {}
    for phase_func in phase_funcs:
        try:
            mask = phase_func(hist)
            if np.any(mask):
                evo_phase[mask] = max_phase
                max_phase += 1
                phase_pass[phase_func] = True
            else:
                phase_pass[phase_func] = False
        except (ValueError, IndexError):
            phase_pass[phase_func] = False
            print(phase_func.__name__, 'failed.')
    return evo_phase, max_phase, phase_pass


def get_evo_stretch_func(hist, xaxis='star_age', phase_funcs=None):
    if phase_funcs is None:
        phase_funcs = [get_pms_mask, get_ms_mask, get_sgb_mask, get_rgb_mask, get_flashes_mask, get_cheb_mask,
                       get_agb_mask]
    evo_phase, max_phase, phase_pass = get_evo_phase(hist, phase_funcs)

    xdata = hist.get(xaxis)

    if np.all(np.diff(xdata) < 0):  # monotonically decreasing
        factor = -1 * xdata
    elif np.all(np.diff(xdata) > 0):  # monotonically increasing
        pass
    else:
        raise ValueError(f'Quantity specified in `xaxis` ({xaxis}) must be monotonic.')

    stretched = np.zeros_like(xdata) + evo_phase
    for phase in range(1, max_phase + 1):
        mask = evo_phase == phase
        if sum(mask) == 0:
            continue
        phase_xdata = xdata[mask]
        xdata_min = min(phase_xdata)
        xdata_max = max(phase_xdata)
        func = ip.interp1d([xdata_min, xdata_max], [0.0, 1.0])

        stretched[mask] += func(phase_xdata)

    func = ip.interp1d(hist.data.model_number, stretched)
    return func, phase_pass, max_phase


def get_bottom_envelope(p, indeces_only=False):
    """

    Args:
        p:
        indeces_only:

    Returns:

    """
    version = str(p.header['version_number'])
    radius = get_radius(p, 'Rsol')
    mass = p.get('mass')
    temperature = p.get('temperature')
    mix_type = p.get('mixing_type')

    if compare_version(version, '15140', '>='):
        prefix = 'post'
    else:
        prefix = 'pre'
    mix_type = np.array(list(map(mix_dict['merged'].get, map(mix_dict[f'{prefix}15140'].get, mix_type))))
    mix_OS = mix_type == mix_dict['merged']['overshoot_mixing']  # OS
    mix_CV = mix_type == mix_dict['merged']['convective_mixing']  # convective
    temp_mask = (temperature[1:] > p.header['Teff'] * 2)
    radiative = False
    try:
        bottom_of_CZ = np.where(mix_CV[:-1] & mix_OS[1:] & temp_mask)[0][0]
        bottom_of_US = np.where(mix_OS[:-1] & (mix_type[1:] != mix_dict['merged']['convective_mixing']) & (mix_type[1:] != mix_dict['merged']['overshoot_mixing']) & temp_mask)[0][0]
    except IndexError:
        radiative = True
        bottom_of_CZ = 0
        bottom_of_US = 0
    if indeces_only:
        return bottom_of_CZ, bottom_of_US
    else:
        if not radiative:
            # grada = p.get('grada')
            # gradr = p.get('gradr')
            # grads = grada[bottom_of_CZ - 10:bottom_of_CZ + 10] - gradr[bottom_of_CZ - 10:bottom_of_CZ + 10]
            # r_bCZ = float(ip.interp1d(grads, radius[bottom_of_CZ - 10:bottom_of_CZ + 10])(0))
            # m_bCZ = float(ip.interp1d(grads, mass[bottom_of_CZ - 10:bottom_of_CZ + 10])(0))
            r_bCZ = radius[bottom_of_CZ]
            m_bCZ = mass[bottom_of_CZ]
        else:
            r_bCZ = radius[bottom_of_CZ]
            m_bCZ = mass[bottom_of_CZ]
    return p.header['model_number'], m_bCZ, r_bCZ, mass[bottom_of_US], radius[bottom_of_US]


def get_lamb2(p, l=1):
    if 'lamb_S2' in p.columns:
        lamb2 = p.data.lamb_S2 * (l * (l + 1) / 2)
    elif 'lamb_Sl1' in p.columns:
        lamb2 = (p.data.lamb_Sl1 / (1e6 / (2 * np.pi))) ** 2 * (l * (l + 1) / 2)  # l part to convert from l=1 to l=l
    else:
        radius = get_radius(p, unit='cm')
        lamb2 = l * (l + 1) * (p.data.csound / radius) ** 2
    return lamb2


def calc_MH(hist, ZX_sol=0.0178, use_mask=None):
    """
    Calculate [M/H] from hist.

    Args:
        hist:
        ZX_sol (float, optional): Solar value for Z/X. Defaults to 0.0178.
        use_mask (bool, np.array, or function, optional): If True, will exclude pre-main sequence.
                If an array of bools will use that as mask. If a function, will call function(hist) and use that as the mask.

    Returns:

    """
    mask = get_mask(hist, use_mask)
    surf_X = hist.data.surface_h1[mask] + hist.data.surface_h2[mask]
    surf_Y = hist.data.surface_he3[mask] + hist.data.surface_he4[mask]
    surf_Z = 1 - surf_X - surf_Y
    FeH = np.log10((surf_Z / surf_X) / ZX_sol)
    return FeH


def calc_deltanu(gs, hist, prefix='profile', suffix='.data.GYRE.sgyre_l', freq_units='uHz'):
    """
    Calculate the large frequency separation Delta nu.

    Args:
        gs (GyreSummary):
        hist (History):
        prefix (str, optional): Part of gyre summary name before the profile number. Defaults to 'profile'.
        suffix (str, optional): Part of gyre summary name after the profile number. Defaults to '.data.GYRE.sgyre_l'.
        freq_units (str): Unit to convert to, must be one of 'uHz', 'mHz', or 'Hz'.

    Returns:
        float: Large frequency separation \\Delta\\nu.
    """
    pnum = int(gs.path.split(prefix)[-1].replace(suffix, ''))
    i_hist = hist.get_profile_index(pnum)[0]
    nu_all = gs.get_frequencies(freq_units)

    mask = gs.data.l == 0
    mask = np.logical_and(mask, gs.data.n_pg > 0)
    nu_max = hist.get('nu_max')[i_hist]
    fsig = (0.66 * nu_max ** 0.88) / 2 / np.sqrt(2 * np.log(2.))
    w = np.exp(-((nu_all[mask][:-1] - nu_max) / fsig) ** 2)
    delta_nus = np.diff(nu_all[mask])
    delta_nu = np.sum(w * delta_nus) / np.sum(w)
    return delta_nu


def calc_deltaPg(gs, hist, l, prefix='profile', suffix='.data.GYRE.sgyre_l'):
    """
    Calculate the period spacing Delta P weighted by the power spectrum envelope.
    Args:
        gs (GyreSummary):
        hist (History):
        l (int): Degree of modes.
        prefix (str, optional): Part of gyre summary name before the profile number. Defaults to 'profile'.
        suffix (str, optional): Part of gyre summary name after the profile number. Defaults to '.data.GYRE.sgyre_l'.

    Returns:
        float: Period spacing Delta P.
    """
    pnum = int(gs.path.split(prefix)[-1].replace(suffix, ''))
    i_hist = hist.get_profile_index(pnum)[0]
    nu_max = hist.get('nu_max')[i_hist]

    if l == 0:
        raise ValueError('Cannot use l=0 for period spacing.')

    mask = gs.data.l == l
    nu = gs.get_frequencies('Hz')[mask]
    dPi = -np.diff(nu ** -1)

    fsig = (0.66 * nu_max ** 0.88) / 2 / np.sqrt(2 * np.log(2.))
    w = np.exp(-((nu[:-1] - nu_max) / fsig) ** 2)
    dPi = np.sum(dPi * w / sum(w))
    return dPi


def correct_seismo(hist, gsspnum, mask, xname='center_he4', do_deltanu=True, do_deltaP=True,
                   prefix='profile', suffix='.data.GYRE.sgyre_l', get_poly=False, weight=True):
    """
    Calculate corrected delta_nu and delta_Pg using GyreSummary instances in gsspnum. The correction is done by
    fitting a 2nd order polynomial to the ratio of delta_nu from the History and from the GyreSummary instances and
    likewise for delta_Pg.

    Args:
        hist (History):
        gsspnum (Tuple of GyreSummary, int): Calculate using ``wsssss.load_data.load_gss(..., return_pnums=True, ...)``.
        mask (bool, np.array, or function, optional): If True, will exclude pre-main sequence.
                If an array of bools will use that as mask. If a function, will call function(hist) and use that as the mask.
        xname (str, optional): Domain to use when fitting.
        do_deltanu (bool, optional): If True, calculate the delta_nu correction. Defaults to True.
        do_deltaP (bool, optional): If True, calculate the delta_Pg correction. Defaults to True.
        prefix (str, optional): Part of gyre summary name before the profile number. Defaults to 'profile'.
        suffix (str, optional): Part of gyre summary name after the profile number. Defaults to '.data.GYRE.sgyre_l'.
        get_poly (bool, optional): If True, returns the correction polynomials instead of corrected values.
        weight (bool, optional): If True, weight points by distance in domain, giving sparse points more weight.

    Returns:
        If do_deltanu is True, returns the corrected delta_nus. If do_deltaP is True, also returns corrected delta_Pg.
        If get_poly is True, returns the the correction polynomials instead of corrected values.
    """
    mask = get_mask(hist, use_mask=mask)

    mnum_min, mnum_max = hist.data.model_number[mask][[0, -1]]

    x = []
    y_nu = []
    y_P = []
    deltanus = []
    deltaPs = []
    for gs, pnum in gsspnum:
        i_hist = hist.get_profile_index(pnum)
        if len(i_hist) == 0:  # pnum not in history
            continue
        i_hist = i_hist[0]
        mnum = hist.data.model_number[i_hist]
        if (mnum < mnum_min) or (mnum > mnum_max):
            continue
        if do_deltanu:
            deltanu = calc_deltanu(gs, hist, prefix=prefix, suffix=suffix)
            deltanus.append(deltanu)
            h_deltanu = hist.data.delta_nu[i_hist]
            y_nu.append(deltanu / h_deltanu)

        if do_deltaP:
            deltaP = calc_deltaPg(gs, hist, 1)
            deltaPs.append(deltaP)
            h_deltaP = hist.data.delta_Pg[i_hist]
            y_P.append(deltaP / h_deltaP)

        x.append(hist.get(xname)[i_hist])
    x = np.array(x)
    y_nu = np.array(y_nu)
    y_P = np.array(y_P)

    i_sort = np.argsort(x)
    x = x[i_sort]
    if weight:
        mid = (x[1:] + x[:-1]) / 2
        w = np.zeros_like(x)
        w[1:-1] = np.diff(mid)
        w[1:-1] = (x[2:] - x[:-2])
        w[0] = 2 * (x[1] - x[0])
        w[-1] = 2 * (x[-1] - x[-2])
        w = np.abs(w)
    else:
        w = np.ones_like(x)

    if do_deltanu:
        y_nu = y_nu[i_sort]
        try:
            nu_poly = np.poly1d(np.polyfit(x, y_nu, 2, w=w))
            nu_f = nu_poly(hist.get(xname)[mask])
            new_deltanu = hist.data.delta_nu[mask] * nu_f
        except TypeError:
            print(hist.path)
            raise
    if do_deltaP:
        y_P = y_P[i_sort]
        try:
            P_poly = np.polyfit(x, y_P, 2, w=w)
        except Exception:
            P_poly = np.array([np.nan] * 3)
        P_poly = np.poly1d(P_poly)
        P_f = P_poly(hist.get(xname)[mask])
        new_deltaP = hist.data.delta_Pg[mask] * P_f

    if do_deltanu and do_deltaP:
        if get_poly:
            return nu_poly, P_poly
        return new_deltanu, new_deltaP
    elif do_deltanu and not do_deltaP:
        if get_poly:
            return nu_poly
        return new_deltanu
    elif not do_deltanu and do_deltaP:
        if get_poly:
            return P_poly
        return new_deltaP


def calc_ModDens(numax, deltanu, deltap):
    """
    Calculate mode density.

    Args:
        numax:
        deltanu:
        deltap:

    Returns:
        Mode density
    """
    return deltanu / (numax ** 2 * 1e-6 * deltap)

def calc_abundance(hist, use_mask=None):
    """
    Calculate surface abundances. Returns abundances (by element, not isotope) and element names.
    Args:
        hist:
        use_mask (bool, np.array, or function, optional): If True, will exclude pre-main sequence.
            If an array of bools will use that as mask. If a function, will call function(hist) and use that as the mask.

    Returns:
        array: Array containing surface abundances.

    """
    mask = get_mask(hist, use_mask)
    cols = [col for col in hist.columns if col.startswith('surface')]
    elems = [re.sub('[0-9]', '', iso.split('_')[1].capitalize()) for iso in cols]
    A = np.zeros((sum(mask), len(np.unique(elems))))
    j = 0
    e0 = elems[0]
    for i, col in enumerate(cols):
        if col == 'surface_neut':
            m = 1
        else:
            e = re.sub('[0-9]', '', col.split('_')[1].capitalize())
            if e != e0:
                j += 1
                e0 = e
            m = float(re.sub('[A-z]', '', col.split('_')[1]))
        A[:, j] += hist.get(col)[mask] / m

    _, idx = np.unique(elems, return_index=True)
    elems = np.array(elems)[np.sort(idx)]

    A = np.log10(A)
    A = A - A[:, np.where(elems == 'H')[0][0]][:, np.newaxis] + 12
    return A, elems



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import pathlib
import shutil
import subprocess
import sys
import time
from argparse import ArgumentParser

import numpy as np
from scipy.integrate import trapezoid

from wsssss import load_data as ld

np.seterr(all='ignore')

if '__file__' not in globals().keys():  # Otherwise doc generation breaks.
    import wsssss
    __file__ = os.path.join(os.path.dirname(wsssss.__file__), '_bin/gyre_driver/gyre_driver.py')

_version = '0.3.0'
_this_dir = pathlib.Path(__file__).parent

# MESA values
pi = 3.1415926535897932384626433832795028841971693993751
Msun = 1.9892E33
Rsun = 6.9598E10
Lsun = 3.8418E33
Teff_sun = 5777E0
G = 6.67428E-8
sigma_b = 5.670400E-5

# From old GYRE_driver
Dnu_sun = 137.
numax_sun = 3100.
nmode = 20


# noinspection PyPep8Naming
def get_nu_max_dnu_dp(args, fpath, l):
    """Calculate nu_max, Delta_nu, and Delta_P for a model."""

    if args.filetype == 'MESA':
        gp = ld.GyreProfile(fpath)
        header = gp.header
        columns = gp.columns
        data = gp.data

        M_star = header['star_mass'] / Msun
        R_star = header['star_radius']
        L_star = header['star_luminosity']
        Teff_star = (L_star / (4 * pi * R_star ** 2 * sigma_b)) ** 0.25
        R_star = R_star / Rsun

        r = data['radius']

        N = np.sqrt(data['brunt_N2'])

    elif args.filetype == 'FGONG':
        raise NotImplementedError('FGONG files not implemented.')
        # header, data = fgong.load_fgong(fpath, return_comment=False)
        #
        # M_star = header[0] / Msun
        # R_star = header[1] / Rsun
        # Teff_star = header[13]
        #
        # r = data[:, 0]
        # m = np.exp(data[:, 1]) * M_star * Msun
        # gamma1 = data[:, 14]
        # gamma1[gamma1 < 0] = 0
        #
        # N = np.sqrt((G * m / r**3) * gamma1)

    elif args.filetype == 'LOSC':
        with open(fpath, 'r') as handle:
            lines = handle.readlines()

        markers = []
        for i, line in enumerate(lines):
            if line.startswith('%'):
                markers.append(i)

        start_data = markers[-2] + 1
        phot_r_cgs, mass_cgs, G_cgs = map(float, lines[start_data].split())

        R_star = phot_r_cgs / Rsun
        M_star = mass_cgs / Msun

        num_zones = int(lines[start_data + 1].strip())
        zone_dat = []
        for i in range(num_zones):
            zone_start = start_data + 2 + i
            zone_dat.append(tuple(map(int, lines[zone_start].split())))

        num_mesh_points = int(lines[zone_start + 1])
        skip_rows = zone_start + 2
        print(fpath)
        r, m_div_r3, P, rho, gamma1, A_div_r = np.loadtxt(str(fpath), skiprows=skip_rows, usecols=[1, 2, 3, 4, 5, 6],
                                                          comments='%').T
        if len(r) != num_mesh_points:
            raise ValueError(f'Loading of {fpath} failed, not enough meshpoints loaded.')

        m_h = 1.67353276e-24
        k_B = 1.38064852e-16
        mu = 1.25  # Assume not very ionized and approx solar composition.
        Teff_star = (P * mu * m_h / (rho * k_B))[-1]

        N_div_r = np.sqrt(-G_cgs * m_div_r3 * A_div_r)
        N = N_div_r * r

    else:
        raise ValueError('Invalid filetype.')

    N_div_r = N / r
    N_div_r[np.isnan(N_div_r)] = 0
    N_div_r[N_div_r < 0] = 0

    nu_max = numax_sun * M_star / (R_star ** 2 * np.sqrt(Teff_star / Teff_sun))
    Dnu = Dnu_sun * np.sqrt(M_star / R_star ** 3)

    DP = np.abs(trapezoid(N_div_r, r))
    DP = 2 * pi ** 2 * (1 / DP) / np.sqrt(l * (l + 1))

    if args.verbose:
        print(f'M_star          = {M_star}')
        print(f'R_star          = {R_star}')
        print(f'Teff            = {Teff_star}')
        print(f'nu_max          = {nu_max}')
        print(f'Dnu             = {Dnu}')
        if l > 0:
            print(f'DP              = {DP}')
        print()

    return nu_max, Dnu, DP


def write_gyre_adin(model_name, l, file_type, suffix, save_modes, grid_type, freq_min, freq_max, n_freq, args):
    """Create the gyre inlist."""

    reduced_name = model_name.name
    gyre_adin = args.gyre_adin_template + reduced_name + f'_l{l}' + suffix
    summary_file = summary_path(model_name, l, suffix, args)
    if args.out_dir == '':
        mode_name_base = model_name
    else:
        mode_name_base = pathlib.Path(args.out_dir) / f'{reduced_name}'

    if type(n_freq) in [int, np.int64]:
        single_scan = True
    else:
        single_scan = False

    if args.verbose:
        print(f'gyre_adin       = {gyre_adin}')
        print(f'l               = {l}')
        print(f'summary_file    = {summary_file}')
        print(f'grid_type       = {grid_type}')
        print(f'file_type       = {file_type}')
        if single_scan:
            print(f'freq_min        = {freq_min}')
            print(f'freq_max        = {freq_max}')
            print(f'n_freq          = {n_freq}')
        else:
            print(f'freq_min        = {repr(freq_min.tolist())}')
            print(f'freq_max        = {repr(freq_max.tolist())}')
            print(f'n_freq          = {repr(n_freq.tolist())}')
        print()

    if args.base_in == '':
        if args.gyre == 'G5':
            base_in = _this_dir / 'INPUT_GYRE_5.2_ad.in'
            base_in_exists = base_in.exists()
        elif args.gyre == 'G4':
            base_in = _this_dir / 'INPUT_GYRE_4.4_ad.in'
            base_in_exists = base_in.exists()
        elif args.gyre == 'G6':
            base_in = _this_dir / 'INPUT_GYRE_6_ad.in'
            base_in_exists = base_in.exists()
        elif args.gyre == 'G7':
            base_in = _this_dir / 'INPUT_GYRE_7_ad.in'
            base_in_exists = base_in.exists()
        elif args.gyre == 'G8':  # Can use the same base inlist
            base_in = _this_dir / 'INPUT_GYRE_8_ad.in'
            base_in_exists = base_in.exists()
        elif args.gyre == 'G9':  # Can use the same base inlist
            base_in = _this_dir / 'INPUT_GYRE_8_ad.in'
            base_in_exists = base_in.exists()
        else:
            raise FileNotFoundError(f'No default base inlist associated with gyre {args.gyre}.')
    else:
        base_in = args.base_in
        base_in_exists = base_in.exists()
    if not base_in_exists:
        raise FileNotFoundError(f'Cannot find a base inlist for gyre. {base_in}')

    shutil.copy2(base_in, gyre_adin)

    if save_modes:
        if file_type == 'LOSC':
            mode_item_list = 'l,n_pg,n_p,freq,E_norm,x,Gamma_1,prop_type,xi_r,xi_h,dE_dx,K'
        elif file_type == 'FGONG':
            mode_item_list = 'l,n_pg,n_p,freq,E_norm,x,m,p,rho,Gamma_1,prop_type,xi_r,xi_h,dE_dx,K'
        elif file_type == 'MESA':
            mode_item_list = 'M_r,l,n_pg,n_p,freq,E_norm,x,m,p,rho,Gamma_1,prop_type,xi_r,xi_h,dE_dx,K'
        else:
            raise ValueError('Invalid filetype.')
        if args.gyre >= 'G5':
            # Gyre 5 uses capital P for pressure and the normalized rotation kernel changed to unnormalized.
            mode_item_list = mode_item_list.replace(',p,', ',P,').replace(',K', ',dbeta_dx')

        mode_output = (f"   mode_template = '{mode_name_base}_l%l_%J_np%p_ng%g.mgyre'\n"
                       f"   mode_file_format = 'TXT'\n"
                       f"   mode_item_list = '{mode_item_list}'"
                       f"/")
        if args.gyre >= 'G7':
            mode_output = mode_output.replace('mode_', 'detail_').replace('%J', '%ID')
    else:
        mode_output = '/'

    freq_units = ("freq_min_units = 'UHZ'\n"
                  "   freq_max_units = 'UHZ'")
    ad_ = 'ad_'
    nad_output = ('&nad_output\n'
                  '/')
    if args.gyre == 'G4':
        freq_units = "freq_units = 'UHZ'"
        ad_ = ''
        nad_output = ''

    if single_scan:
        scan_str = (f"&scan\n"
                    f"   grid_type = '{grid_type}'\n"
                    f"   {freq_units}\n"
                    f"   freq_min =   {freq_min}\n"
                    f"   freq_max =   {freq_max}\n"
                    f"   n_freq =     {n_freq}\n"
                    f"/")
    else:
        num_scan = len(n_freq)
        scan_str = ''
        for i in range(num_scan):
            freq_min_i = freq_min[i]
            freq_max_i = freq_max[i]
            n_freq_i = n_freq[i]
            scan_str += (f"&scan\n"
                         f"   grid_type = '{grid_type}'\n"
                         f"   {freq_units}\n"
                         f"   freq_min =   {freq_min_i}\n"
                         f"   freq_max =   {freq_max_i}\n"
                         f"   n_freq =     {n_freq_i}\n"
                         f"/\n")

    s = (f"\n"
         f"! Generated by py_gyre_driver {_version}.\n\n"
         f"&model\n"
         f"   model_type = 'EVOL'\n"
         f"   file = '{model_name}'\n"
         f"   file_format = '{file_type}'\n"
         f"/\n"
         f"&{ad_}output\n"
         f"   summary_file = '{summary_file}'\n"
         f"   summary_file_format = 'TXT'\n"
         f"   summary_item_list = '{args.summary_item_list}'\n"
         f"   freq_units = 'UHZ'\n"
         f"{mode_output}\n"
         f"{nad_output}\n"
         f"{scan_str}"
         f"&mode\n"
         f"   l = {l}\n"
         f"/\n")

    with open(gyre_adin, 'a') as handle:
        handle.write(s)

    return


def run_gyre(model_name, l, suffix, args):
    """Run gyre an instance of gyre."""
    reduced_name = model_name.name
    gyre_adin = args.gyre_adin_template + reduced_name + f'_l{l}' + suffix
    summary_file = summary_path(model_name, l, suffix, args)

    gyre_exec = get_gyre(args, True)[0]
    if args.no_output:
        cmd_end = ' > /dev/null 2>& 1'
        pipe = None
    else:
        cmd_end = ''
        pipe = subprocess.PIPE

    if args.verbose:
        print(os.getcwd())
        print(f'Calling {gyre_exec} {gyre_adin}\n')

    if os.path.exists(summary_file):
        os.remove(summary_file)

    if not args.skip_calc or (args.skip_calc and l == 0):
        if args.source != '':
            output = subprocess.run(f'source {args.source}; {gyre_exec} {gyre_adin} {cmd_end}',
                                    shell=True, executable="/bin/bash", stderr=pipe)
        else:
            output = subprocess.run(f'{gyre_exec} {gyre_adin} {cmd_end}',
                                    shell=True, executable="/bin/bash", stderr=pipe)

        if not args.no_output:
            if not output.stderr == b'':
                if "ASSERT 'k == k_chk' failed at line 303" in output.stderr.decode():
                    raise ValueError(f"Gyre 4.4 does not support the version of MESA used to generate {model_name}")

                raise ChildProcessError(f'The following command encountered an error:\n.'
                                        f'{output.args}\n\n'
                                        f'{output.stderr.decode()}\n')
    else:
        output = f'skipped {gyre_exec} {gyre_adin}'

    return output


def get_gyre(args, check, print_warning=False):
    """Get the path to the gyre executable and check if it is the correct version."""
    version = args.gyre
    environ = os.environ

    if f'GYRE_DIR_{version}' in environ:
        path = pathlib.Path(environ[f'GYRE_DIR_{version}'])
    else:
        path = pathlib.Path(environ['GYRE_DIR'])

        if check:  # Check in $GYRE_DIR/src/common/gyre_version.fpp what version of gyre is found.
            version_file = path / 'src' / 'common' / 'gyre_version.fpp'
            if not version_file.exists():  # Different file name for 7.2 and after
                version_file = path / 'src' / 'common' / 'version_m.fypp'

            with open(version_file, 'r') as handle:
                lines = handle.readlines()

            version_str = ''
            for line in lines:
                if 'VERSION = ' in line.upper():
                    line = line.upper()
                    version_str = line.split('VERSION =')[1].strip().replace("'", '').replace('(', '').replace(')', '')
                    break

            if version_str == '':
                raise ValueError(f'Could not find gyre version of $GYRE_DIR in file\n'
                                 f'{version_file}')

            if not version[1] == version_str[0]:
                if args.lenient:
                    if print_warning:
                        print('###################################################################')
                        print(f'Could not find the required gyre version `{version}` in $GYRE_DIR.\n'
                              f'Using gyre version {version_str} instead.')
                        print('###################################################################')
                        print()
                    version = f'G{version_str[0]}'
                else:
                    raise ValueError(f'Could not find the required gyre version `{version}` in $GYRE_DIR.\n'
                                     f'Found gyre version {version_str} in `{path}`.')

    if version <= 'G4':
        path = path / 'bin' / 'gyre_ad'
    else:
        path = path / 'bin' / 'gyre'

    return path, version


def merge_summary_parts(model_name, l, num_scan, keep_files=False):
    """Append all partial runs for a single `l` into one summary file."""

    all_lines = []
    summary_files = []
    set_header = False
    for i in range(num_scan):
        partial = f'.part{i + 1}of{num_scan}'
        summary_file = summary_path(model_name, l, partial, args)
        if not summary_file.exists():  # Skip summary files which found no modes.
            continue

        summary_files.append(summary_file)
        with open(summary_file, 'r') as handle:
            lines = handle.readlines()

        if not set_header:
            all_lines += lines
            set_header = True
        else:
            all_lines += lines[6:]

    new_summary_file = summary_path(model_name, l, '', args)

    with open(new_summary_file, 'w') as handle:
        handle.writelines(all_lines)

    if not keep_files:
        for summary_file in summary_files:
            os.remove(summary_file)


def merge_summary_l(model_name, args, keep_files=False):
    """Append all runs for all `l` into one summary file."""

    all_lines = []
    summary_files = []
    set_header = False
    for l in args.ll:
        summary_file = summary_path(model_name, l, '', args)
        if not summary_file.exists():  # Skip summary files which found no modes.
            continue

        summary_files.append(summary_file)
        with open(summary_file, 'r') as handle:
            lines = handle.readlines()

        if not set_header:
            all_lines += lines
            set_header = True
        else:
            all_lines += lines[6:]

    new_summary_file = summary_path(model_name, '', '', args)
    new_summary_file = pathlib.Path(str(new_summary_file).replace('.sgyre_l', args.summary_suffix))
    with open(new_summary_file, 'w') as handle:
        handle.writelines(all_lines)

    if not keep_files:
        for summary_file in summary_files:
            os.remove(summary_file)


def summary_path(model_name, l, suffix, args):
    reduced_name = model_name.name

    if args.out_dir == '':
        summary_file = pathlib.Path(f'{model_name}.sgyre_l{l}{suffix}')
    else:
        summary_file = pathlib.Path(args.out_dir) / f'{reduced_name}.sgyre_l{l}{suffix}'

    return summary_file


# noinspection PyPep8Naming
def calc_scan(model_name, l, args):
    nu_max, Dnu, DP = get_nu_max_dnu_dp(args, model_name, l)
    n_sig_lo = args.n_sig_lo
    n_sig_hi = args.n_sig_hi

    fmid = nu_max
    fsig = (0.66 * nu_max ** 0.88) / 2 / np.sqrt(2 * np.log(2.))  # Mosser 2012a

    fmin0 = max(1e-4, fmid - n_sig_lo * fsig)
    fmax0 = fmid + n_sig_hi * fsig

    n_freqDnu = int(np.ceil((fmax0 - fmin0) / Dnu))

    if l == 0:
        n_freq = 3 * n_freqDnu
        num_scan = 1
        fmin = fmin0
        fmax = fmax0
    else:
        l0_summary_file = summary_path(model_name, 0, '', args)
        gs = ld.GyreSummary(l0_summary_file)
        freqs_l0 = gs.get('Re(freq)')[gs.get('l') == 0]

        n_freqDP = np.ceil((fmax0 - fmin0) / (1e6 / (1e6 / fmin0 - DP) - fmin0)).astype(int)

        if n_freqDnu > n_freqDP:
            n_freqDnu = np.ceil((fmax0 - fmin0) / Dnu)
            n_freq = int(5 * n_freqDnu)
            num_scan = 1
            fmin = fmin0
            fmax = fmax0

        else:
            fmin = np.concatenate([[fmin0], freqs_l0[:-1]])
            fmax = np.concatenate([freqs_l0[1:], [fmax0]])

            if args.pmode:
                # See figure 4 in https://arxiv.org/abs/1108.4777 for origin of 0.13.
                if l % 2 == 0:  # even l
                    fmid = fmax - 0.13 * (fmax - fmin)
                if l % 2 == 1:  # odd l
                    fmid = np.mean([fmin, fmax], axis=0)

                fsig = np.maximum(0.15, np.minimum(0.5 * (fmax - fmin), nmode * (1e6 / (1e6 / fmid - DP) - fmid)))
                fmin = fmid - fsig
                fmax = fmid + fsig

                # Don't search outside l=0 mode frequencies
                fmin = np.maximum(fmin, freqs_l0[0])
                fmax = np.minimum(fmax, freqs_l0[-1])

            n_freq = 5 * np.maximum(5, np.ceil((fmax - fmin) / (1e6 / (1e6 / fmin - DP) - fmin)).astype(int))

            num_scan = len(fmin)
    n_freq = n_freq * args.f_nfreq
    return fmin, fmax, n_freq, num_scan, nu_max


def split_scan(fmin, fmax, n_freq, grid_type, args):
    if grid_type == 'LINEAR':

        fmin_new = []
        fmax_new = []
        n_freq_new = []
        for i, n in enumerate(n_freq):
            if n <= args.batch:
                fmin_new.append(fmin[i])
                fmax_new.append(fmax[i])
                n_freq_new.append(n)
            else:
                num_split = n // args.batch + 1
                sub_intervals = np.linspace(fmin[i], fmax[i], num_split + 1)
                sub_fmin = sub_intervals[:-1]
                sub_fmax = sub_intervals[1:]
                sub_n = (int(n / num_split) + 1) * np.ones_like(sub_fmin, dtype=int)

                fmin_new.extend(sub_fmin)
                fmax_new.extend(sub_fmax)
                n_freq_new.extend(sub_n)
        fmin = np.array(fmin_new)
        fmax = np.array(fmax_new)
        n_freq = np.array(n_freq_new, dtype=int)

    else:
        raise NotImplementedError(f'Only `grid_type = LINEAR` is implemented.')

    num_scan = len(n_freq)

    return fmin, fmax, n_freq, num_scan


def do_gyre_sim(fpath, args):
    num_scan = 0
    try:
        for l in args.ll:
            if l == 0:
                grid_type = 'LINEAR'
            else:
                grid_type = 'INVERSE'

            fmin, fmax, n_freq, num_scan, nu_max = calc_scan(fpath, l, args)
            if (l > 0) and (nu_max < args.min_numax):
                print(f'Skipping {fpath} l={l} as numax {nu_max} below min-numax {args.min_numax}.')
                continue

            # Split each frequency scan into a different gyre run. Usually is a bit faster.
            if num_scan > 1 and args.parts:
                # Split into scans upto size `args.batch`. Can be slow if batch is too small.
                if args.batch is not None and any(n_freq > args.batch):
                    fmin, fmax, n_freq, num_scan = split_scan(fmin, fmax, n_freq, grid_type, args)

                for i in range(len(n_freq)):
                    fmin_i = fmin[i]
                    fmax_i = fmax[i]
                    n_freq_i = n_freq[i]

                    part_suffix = f'.part{i + 1}of{num_scan}'

                    write_gyre_adin(fpath, l, args.filetype, part_suffix, args.save_modes,
                                    grid_type, fmin_i, fmax_i, n_freq_i, args)
                    run_gyre(fpath, l, part_suffix, args)

            else:
                write_gyre_adin(fpath, l, args.filetype, '', args.save_modes, grid_type, fmin, fmax, n_freq, args)
                run_gyre(fpath, l, '', args)

            if args.parts and num_scan > 1:
                merge_summary_parts(fpath, l, num_scan)
        if not args.no_merge:
            merge_summary_l(fpath, args)
    except KeyboardInterrupt:
        print('Stopping GYRE')
        if num_scan > 1 and args.parts:
            merge_summary_parts(fpath, l, num_scan, keep_files=True)  # Merge the summaries for the interrupted l.
        merge_summary_l(fpath, args, keep_files=True)
        raise


def sort_files(files):
    path_lengths = np.array([a for a in enumerate(map(lambda x: len(str(x)), files))])
    i_sort = np.argsort(path_lengths[:, 1])
    files = np.asarray(files, dtype=object)[i_sort]
    path_lengths = path_lengths[i_sort]

    lengths = list(set(path_lengths[:, 1]))
    lengths.sort()
    for length in lengths:
        i_sub = np.where(path_lengths[:, 1] == length)[0]
        files[i_sub] = files[i_sub][np.argsort(files[i_sub])]

    files = list(files)
    return files


def check_args(args):
    if args.gyre >= 'G6':
        parts_set = False
        if args.parts:
            parts_set = True
        batch_set = False
        if args.batch is not None:
            batch_set = True

        if parts_set or batch_set:
            print("When using gyre 6 or above --parts and/or --batch are no longer required.")
            print("Turning off --parts and --batch")
            args.parts = False
            args.batch = None

    if args.ll.startswith('mode'):
        args.save_modes = True
        args.ll = args.ll.replace('mode', '')

    args.ll = [int(l) for l in args.ll]
    args.ll.sort()

    if 0 not in args.ll:
        args.ll = [0] + args.ll  # Need l=0 modes to estimate frequency scans.

    if type(args.files) == list:
        args.files = [pathlib.Path(fpath.strip()) for fpath in args.files]
    else:
        args.files = [pathlib.Path(args.files.strip())]

    if args.skip_existing:
        existing_out_files = list(pathlib.Path(args.out_dir).glob('*.sgyre_l'))
        existing_out_files = [_.absolute() for _ in existing_out_files]
        do_files = [model_name for model_name in args.files if
                    summary_path(model_name, '', '', args).absolute() not in existing_out_files]
        if args.verbose:
            print(f'num_prof_skipped= {len(args.files) - len(do_files)}')
        args.files = do_files

    # Sort files by length, then by filename.
    if not args.nosort and len(args.files) > 1:
        args.files = sort_files(args.files)

    # Create output directory
    if args.out_dir != '':
        args.out_dir = pathlib.Path(args.out_dir)
        args.out_dir.mkdir(exist_ok=True)

    # Create and/or clean input directory
    if args.in_dir != '':
        gyre_adin_template = args.in_dir + '/gyre_ad.in_'
    else:
        args.in_dir = 'gyre_ad.in'
        gyre_adin_template = 'gyre_ad.in/gyre_ad.in_'
    args.in_dir = pathlib.Path(args.in_dir)
    args.in_dir.mkdir(exist_ok=True)

    if args.base_in != '':
        args.base_in = pathlib.Path(args.base_in)

    # If --lenient, switch to the version of gyre found.
    original_gyre = args.gyre
    checked_gyre = get_gyre(args, True, True)[1]
    args.gyre = checked_gyre

    args.gyre_adin_template = gyre_adin_template
    args.original_gyre = original_gyre
    return args


def get_parser():
    parser = ArgumentParser(description="""
    This program will run GYRE, estimating appropriate number of scan points for different angular orders l
    and range of frequency.
    """)

    parser.add_argument('ll', type=str,
                        help='Degrees of modes to scan. eg. `012` or `mode0`. '
                             'Including `mode` overrides `--save-modes`.')
    parser.add_argument('filetype', type=str, choices=['FGONG', 'MESA', 'LOSC'], default='MESA',
                        help='Filetype of profiles that gyre will read.')
    parser.add_argument('files', type=str, nargs='*',
                        help='Paths to profile files for gyre to use.')
    parser.add_argument('--gyre', type=str, choices=['G4', 'G5', 'G6', 'G7', 'G8', 'G9'], default='G9',
                        help='Which version of gyre to use.')
    parser.add_argument('--pmode', action='store_const', const=True, default=False,
                        help='If set, scan for modes around the expected frequencies.')
    parser.add_argument('--save-modes', action='store_const', const=True, default=False,
                        help='Save mode profiles.')
    parser.add_argument('--verbose', '-v', action='store_const', const=True, default=False,
                        help='Print info about gyre runs.')
    parser.add_argument('--parts', '-p', action='store_const', const=True, default=False,
                        help='Run each frequency scan window individually. This is probably a bit faster.')
    parser.add_argument('--source', type=str, default='',
                        help='Source this file before running gyre.')
    parser.add_argument('--lenient', action='store_const', const=True, default=False,
                        help='If set, will not stop `gyre_driver` if the version of gyre given to `--gyre` is different'
                             ' than the version found in `$GYRE_DIR`.')
    parser.add_argument('--out-dir', type=str, default='gyre_out',
                        help='Directory in which to save output.')
    parser.add_argument('--in-dir', type=str, default='gyre_in',
                        help='Directory in which to save inlist files.')
    parser.add_argument('--base-in', type=str, default='',
                        help='Path to the base inlist to use.')
    parser.add_argument('--nosort', action='store_const', const=True, default=False,
                        help="Don't sort input files by profile number.")
    parser.add_argument('--no-merge', '-m', action='store_const', const=True, default=False,
                        help="Don't merge the final summary files into a single file.")
    parser.add_argument('--batch', type=int, default=None,
                        help='Split scans into batches of up to N frequencies in order to'
                             ' get around memory issues with broken prune_modes in 5.2.')
    parser.add_argument('--skip-calc', action='store_const', const=True, default=False,
                        help='Do everything except run gyre for l >= 1. Useful to only generate inlists.')
    parser.add_argument('--skip-existing', action='store_const', const=True, default=False,
                        help='Skip running gyre for existing runs. Only works if summary files are merged.')
    parser.add_argument('--min-numax', type=float, default=0,
                        help='Models with numax in uHz lower than this will only calculate l=0 modes.')
    parser.add_argument('--version', action='version', version=f'gyre_driver {_version}')
    parser.add_argument('--summary-item-list', type=str, default='l,n_pg,n_p,n_g,freq,E_norm,M_star,R_star,L_star,E',
                        help='Summary item list for gyre.')
    parser.add_argument('--n-sig-lo', type=float, default=2,
                        help="Number of power envelope sigma to look for modes below numax.")
    parser.add_argument('--n-sig-hi', type=float, default=2,
                        help="Number of power envelope sigma to look for modes above numax.")
    parser.add_argument('--f-nfreq', type=int, default=1,
                        help='Factor to increase number of scan frequencies.')
    parser.add_argument('--summary-suffix', type=str, default='.sgyre_l',
                        help='Merged summary file suffix.')
    parser.add_argument('--no-output', action='store_const', const=True, default=False,
                        help='If set, pipe all terminal output to /dev/null.')
    return parser


def run(args=None):
    t_start = time.time()
    cwd = os.getcwd()

    if args is None:
        parser = get_parser()
        args = parser.parse_args()

    if not args.nosort and len(args.files) > 1:
        args.files = sort_files(args.files)

    if args.filetype == 'LOSC':
        print('Warning! LOSC does not supply an effective temperature, will use an estimate instead.')

    if args.verbose:
        print()
        print('Parsed inputs:')
        print(f'll              = {args.ll}')
        print(f'filetype        = {args.filetype}')
        print(f'files           = {" ".join([str(_) for _ in args.files])}')
        print(f'--gyre          = {args.gyre}')
        print(f'--pmode         = {args.pmode}')
        print(f'--save-modes    = {args.save_modes}')
        print(f'--verbose       = {args.verbose}')
        print(f'--parts         = {args.parts}')
        print(f'--source        = {args.source}')
        print(f'--lenient       = {args.lenient}')
        print(f'--out-dir       = {args.out_dir}')
        print(f'--in-dir        = {args.in_dir}')
        print(f'--base-in       = {args.base_in}')
        print(f'--nosort        = {args.nosort}')
        print(f'--no-merge      = {args.no_merge}')
        print(f'--batch         = {args.batch}')
        print(f'--skip-calc     = {args.skip_calc}')
        print(f'--skip-existing = {args.skip_existing}')
        print(f'--n-sig-lo      = {args.n_sig_lo}')
        print(f'--n-sig-hi      = {args.n_sig_hi}')
        print(f'--f-nfreq      = {args.f_nfreq}')
        print(f'--summary-suffix= {args.summary_suffix}')
        print()

    # Load environment variables from --source and keep a copy of the old ones.
    environ_original = os.environ.copy()
    if args.source != '':
        env_out = subprocess.check_output(f"source {args.source}; env -0", shell=True, executable="/bin/bash")
        environ_used = dict(line.decode().partition('=')[::2] for line in env_out.split(b'\0'))
        if '' in environ_used:
            environ_used.pop('')
        os.environ.clear()
        os.environ.update(environ_used)
    environ_used = os.environ.copy()

    args = check_args(args)

    if args.verbose:
        print('Processed inputs:')
        print(f'll              = {args.ll}')
        print(f'files           = {" ".join([str(_) for _ in args.files])}')
        print(f'--save-modes    = {args.save_modes}')
        print(f'--gyre          = {args.gyre}')
        print(f'--out-dir       = {args.out_dir}')
        print(f'--in-dir        = {args.in_dir}')
        print(f'--base-in       = {args.base_in}')
        print()

    for file in args.files:
        do_gyre_sim(file, args)

    t_end = time.time()
    t_taken = t_end - t_start
    if not args.no_output:
        print(f"Total time taken: {int(t_taken // 3600)}h{int(t_taken // 60) % 60}m{t_taken % 60 :.2f}s\n")

    if args.original_gyre != args.gyre:
        print('###################################################################')
        print(f'Could not find the required gyre version `{args.original_gyre}` in $GYRE_DIR.\n'
              f'Used gyre version `{args.gyre}` instead.')
        print('###################################################################')
        print()

    # Reset to the old environment variables.
    os.environ.clear()
    os.environ.update(environ_original)
    return 0

if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

import f90nml
import numpy as np

from ..functions import get_mesa_version, compare_version
from ..constants import post15140 as const

parser = f90nml.Parser()
parser.global_start_index = 1

# Locations of *.defaults files in $MESA_DIR.
defaults = {'star_job': 'star/defaults/star_job.defaults',
            'eos': 'eos/defaults/eos.defaults',
            'kap': 'kap/defaults/kap.defaults',
            'controls': 'star/defaults/controls.defaults',
            'pgstar': 'star/defaults/pgstar.defaults',
            'astero': ['star/astero/defaults/astero_search.defaults', 'astero/defaults/astero_search.defaults']
            }


def inlist_diff(dict1, dict2):
    """
    Compare two inlists and show which items change or remain the same.

    Args:
        dict1 (dict):
        dict2 (dict):

    Returns:
        dict: Items which remain the same are in ``result['same']``,
            and those which changed in ``result['changed']``.
    """
    keys1 = list(dict1.keys())
    keys2 = list(dict2.keys())

    unique_keys = []
    for k in keys1 + keys2:
        if k not in unique_keys:
            unique_keys.append(k)

    result = {'same': {}, 'changed': {}}

    for k in unique_keys:
        v1 = None
        v2 = None

        if k in keys1:
            v1 = dict1[k]
        if k in keys2:
            v2 = dict2[k]

        if isinstance(v1, (f90nml.Namelist, dict)) or isinstance(v2, (f90nml.Namelist, dict)):
            raise ValueError(f'Cannot compare nested dicts for key {k}.')

        if v1 == v2:
            result['same'][k] = v1
        else:
            result['changed'][k] = (v1, v2)

    return result


def evaluate_inlist_str(inlist_str, inlist_dir):
    """
    Fully evaluate `inlist_str` as if it were written in `inlist_dir`, following all ``read_extra_*_inlist`` if
    they are ``.true..``.

    Args:
        inlist_str (str): `str` form of an inlist.
        inlist_dir (str): Reference directory for reading other files.

    Returns:
        dict: Evaluated inlist.

    """
    inlist = parser.reads(inlist_str)
    return _evaluate_inlist(inlist, inlist_dir)


def evaluate_inlist(path):
    """
    Fully evaluate the inlist at `path`, following all ``read_extra_*_inlist`` if they are ``.true..``.
    Args:
        path (str): Path to `inlist`.

    Returns:
        dict: Evaluated inlist.

    """
    inlist = parser.read(path)
    inlist_dir = os.path.dirname(path)
    return _evaluate_inlist(inlist, inlist_dir)


def get_inlist_type(inlist):
    key_type = 'old'
    for kind in defaults.keys():
        if kind not in inlist.keys():
            continue

        for key, item in list(inlist[kind].items()):
            if isinstance(item, list):
                key_type = 'new'
                return key_type
    return key_type

def _evaluate_inlist(inlist, inlist_dir):
    """
    Fully evaluate `inlist` as if it were written in `inlist_dir`, following all ``read_extra_*_inlist`` if
    they are ``.true..``.

    Args:
        inlist (dict): Inlist dictionary.
        inlist_dir (str): Directory which contains `inlist`.

    Returns:
        dict: Evaluated inlist.

    """
    out = {}
    key_type = 'old'
    for kind in defaults.keys():
        if kind not in inlist.keys():
            continue

        for key, item in list(inlist[kind].items()):
            if isinstance(item, list):
                key_type = 'new'
                inlist[kind].pop(key)
                for i in range(len(item)):
                    if item[i] is not None:
                        inlist[kind][f'{key}({i + 1})'] = item[i]

        out[kind] = dict(inlist[kind])

    if key_type == 'old':
        max_extra_inlists = 5  # Hard coded in MESA
        read_extra = 'read_extra_{}_inlist{}'
        read_extra_filename = 'extra_{}_inlist{}_name'
    else:
        max_extra_inlists = const.max_extra_inlists
        read_extra = 'read_extra_{}_inlist({})'
        read_extra_filename = 'extra_{}_inlist_name({})'

    to_read = {}
    for kind in defaults.keys():
        if kind not in inlist.keys():
            continue
        to_read[kind] = []

        for i in range(max_extra_inlists):
            key = read_extra.format(kind, i+1)
            if key in inlist[kind].keys():
                if inlist[kind][key]:
                    to_read[kind].append(inlist[kind][read_extra_filename.format(kind, i+1)])

    inlists = {}
    for kind, fnames in to_read.items():
        inlists[kind] = []
        for fname in fnames:
            sub_inlist = parser.read(os.path.join(inlist_dir, fname))
            sub_inlist = _evaluate_inlist(sub_inlist, inlist_dir)

            if kind in sub_inlist.keys():
                inlists[kind].append(sub_inlist[kind])

    for kind, sub_inlists in inlists.items():
        if kind not in inlist.keys():
            out[kind] = {}
        for sub_inlist in sub_inlists:
            out[kind].update(sub_inlist)

    for kind in out.keys():
        for key, value in list(out[kind].items()):  # Consume whole generator as we need to change the underlying dict.
            if '_inlist' in key:  # Don't want read_extra inlist keys anymore.
                _ = out[kind].pop(key)
    return out


def variable_to_string(value):
    """
    Convert `value` to a fortran-compatible string. Can be a `string`, `bool`, `int`, or `float`.

    Args:
        value (str, bool, int, or float): Value to convert.

    Returns:
        str: Fortran-compatible string.
    """

    if type(value) == bool:
        parsed = f'.{str(value).lower()}.'
    elif type(value) == str:
        parsed = f"'{value}'"
    else:
        if isinstance(value, (int, np.int_)):
            parsed = str(value)
        else:
            parsed = f'{value:g}'.replace('e', 'd').replace('d+', 'd')
            if 'd' not in parsed:
                parsed += 'd0'
    return parsed


def write_inlist(inlist, path, header='', mode='w'):
    """
    Write the `inlist` to `path`. `header` is prepended to the string which is written. `mode` is passed to ``open``.

    Args:
        inlist (dict):
        path (str): Inlist file path.
        header (str):
        mode (str): File mode passed to ``open``.

    """
    if header and not header.startswith('!'):
        header = '!' + header
    s = '' + header

    for name in inlist.keys():
        s += f'&{name} ! start\n'
        for key, value in inlist[name].items():
            s += f'    {key} = {variable_to_string(value)}\n'

        s += f'/ !{name} end\n'

    with open(path, mode) as handle:
        handle.write(s)

def print_dict(to_print):
    """
    Print the dict `to_print` with a key and value on every line.

    Args:
        to_print (dict):
    """
    for k, v in to_print.items():
        print(f'{k}: {v}')


def load_MESA_defaults(mesa_dir):
    mesa_version = get_mesa_version(mesa_dir)
    if compare_version(mesa_version, 'r23.05.1', '<'):
        max_extra_inlists = 5  # Hard coded in MESA
        read_extra = 'read_extra_{}_inlist{}'
        read_extra_filename = 'extra_{}_inlist{}_name'
    else:
        max_extra_inlists = const.max_extra_inlists
        read_extra = 'read_extra_{}_inlist({})'
        read_extra_filename = 'extra_{}_inlist_name({})'

    if compare_version(mesa_version, '15140', '>='):
        astero_index = 1
    else:
        astero_index = 0

    inl_str = ''
    default_keys = get_mesa_defaults(mesa_dir)
    for namelist in default_keys.keys():
        if namelist == 'astero':
            inl_str += (f"&{namelist}\n"
                        f"   {read_extra.format(namelist, 1)} = .true.\n"
                        f"   {read_extra_filename.format(namelist, 1)} = '{mesa_dir}/{defaults['astero'][astero_index]}'\n"
                        f"/\n\n")
        else:
            inl_str += (f"&{namelist}\n"
                        f"   {read_extra.format(namelist, 1)} = .true.\n"
                        f"   {read_extra_filename.format(namelist, 1)} = '{mesa_dir}/{defaults[namelist]}'\n"
                        f"/\n\n")
    print(inl_str)
    inlist_defaults = evaluate_inlist_str(inl_str, '')
    return inlist_defaults

def compare_inlist(path1, path2, show_same=False):
    inlist1 = evaluate_inlist(path1)
    inlist2 = evaluate_inlist(path2)

    star_job = inlist_diff(inlist1['star_job'], inlist2['star_job'])
    controls = inlist_diff(inlist1['controls'], inlist2['controls'])

    print(f'left : {path1}')
    print(f'right: {path2}')
    print('')
    print('####### star_job differences #######')
    print_dict(star_job['changed'])
    if show_same:
        print('####### star_job same #######')
        print_dict(star_job['same'])
    print('')

    print('####### controls differences #######')
    print_dict(controls['changed'])
    if show_same:
        print('####### controls same #######')
        print_dict(controls['same'])

def get_mesa_defaults(mesa_dir):
    """
    Get the available MESA inlist options for the version of MESA installed in `mesa_dir`.

    Args:
        mesa_dir (str): ``$``MESA_DIR`` root directory.

    Returns:
        dict: List of keys for each inlist type (e.g. ``star_job``, ``controls``, etc.).
    """
    version = get_mesa_version(mesa_dir)
    if compare_version(version, '15140', '>='):
        namelists = ('star_job', 'eos', 'kap', 'controls', 'astero', 'pgstar')
        separate_eoskap = True
        astero_index = 1
    else:
        namelists = ('star_job', 'controls', 'pgstar')
        separate_eoskap = False
        astero_index = 0

    if compare_version(version, 'r23.05.1', '>='):
        extra_inlist_as_list = True
    else:
        extra_inlist_as_list = False


    keys_defaults = {}
    for nml_type in namelists:
        if nml_type == 'astero':
            path_to_file = os.path.join(mesa_dir, defaults[nml_type][astero_index])
        else:
            path_to_file = os.path.join(mesa_dir, defaults[nml_type])

        with open(path_to_file, 'r') as handle:
            s_raw = handle.read()

        # Get rid of empty lines and comments
        s = s_raw.lower().splitlines()
        s = [_ for _ in s if not _.strip().startswith('!')]  # Get rid of fortran comments
        s = [_ for _ in s if _.strip()]  # Get rid of ''
        s_no_com = '\n'.join(s)

        keys = [_.split('=', maxsplit=1)[0].strip() for _ in s]
        values = [_.split('=', maxsplit=1)[1].split('!')[0].strip() for _ in s]
        keys_defaults[nml_type] = keys
    return keys_defaults


def check_inlist(path, mesa_dir):
    """
    Check whether the options in the inlist at `path` are available in the version of mesa specified in `mesa_dir`.
    Will try to detect if the options have been moved to ``eos`` or ``kap`` from ``star_job`` or ``controls``.

    Args:
        path (str): Path to an inlist.
        mesa_dir (str): ``$``MESA_DIR`` root directory.

    Returns:
        dict: For each inlist option, states whether it is available in the version of mesa specified in `mesa_dir`.
    """
    version = get_mesa_version(mesa_dir)
    if compare_version(version, '15140', '>='):
        separate_eoskap = True
    else:
        separate_eoskap = False

    nml = evaluate_inlist(path)
    keys_defaults = get_mesa_defaults(mesa_dir)

    all_checked = {}
    for nml_type, sub_nml in nml.items():
        if nml_type not in keys_defaults.keys():
            all_checked[nml_type] = {f'!namelist {nml_type}': 'not in defaults'}
            continue
        key_arr = [key for key in sub_nml.keys()]

        def_key = keys_defaults[nml_type]

        in_default = {key: key in def_key for key in key_arr}

        in_default = {}
        for key in key_arr:
            key_to_check = key
            if nml_type == 'controls':
                if 'ctrl(' in key:
                    key_to_check = key.split('(')[0] + '(1:num_x_ctrls)'
            in_default[key] = key_to_check in def_key

        to_remove = [key for key in key_arr if not in_default[key]]

        orders = {key: 'not in defaults' for key in to_remove}
        # Check if have to move from star_job/controls to eos or kap
        if (nml_type in ['star_job', 'controls']) and separate_eoskap:
            for key in to_remove:
                for new_nml_type in ['eos', 'kap']:
                    if key in keys_defaults[new_nml_type]:
                        orders[key] = f'move to {new_nml_type}'
                        break
                    else:
                        if key.startswith('kappa'):
                            new_key = 'kap' + key[5:]
                            if new_key in keys_defaults[new_nml_type]:
                                orders[key] = f'rename to {new_key} and move to {new_nml_type}'
        all_checked[nml_type] = orders

    # Print nicely
    for nml_type, orders in all_checked.items():
        if len(orders) == 0:
            continue
        print(nml_type, len(orders))
        for key, order in orders.items():
            print(f'    {key}: {order}')
        print()

    return all_checked



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# TODO: check environment for $MESA_DIR and get constants from there, if this version and a history/profile have a
# TODO: mismatch raise an error.

from . import functions
from . import load_data
from . import plotting
from . import inlists
import importlib.metadata

__all__ = ['functions', 'load_data', 'plotting', 'inlists']
__version__ = importlib.metadata.version(__package__ or __name__)
del importlib



#/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import profile
import unittest
import numpy as np

from wsssss import load_data as ld
from wsssss import functions as uf

from .common import have_mesa_data

have_mesa_data()

must_have_environ = ['MESA_DIR']
for env in must_have_environ:
    if env not in os.environ:
        raise EnvironmentError(f'{env} not set.')

test_data = os.path.join(os.path.dirname(__file__), 'data', 'mesa')

class TestFunctions(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.hist = ld.History(f'{test_data}/0000/LOGS/history.data')

    def test_get_mesa_version(self):
        mesa_version = uf.get_mesa_version(os.environ['MESA_DIR'])

    def test_compare_mesa_version(self):
        self.assertTrue(uf.compare_version('11701', 'r24.03.1', '<'))
        self.assertTrue(uf.compare_version('11701', 'r24.03.1', '<='))
        self.assertFalse(uf.compare_version('11701', 'r24.03.1', '>'))
        self.assertFalse(uf.compare_version('11701', 'r24.03.1', '>='))
        self.assertFalse(uf.compare_version('11701', 'r24.03.1', '=='))
        self.assertTrue(uf.compare_version('11701', 'r24.03.1', '!='))
        self.assertTrue(uf.compare_version('8888', '11701', '<'))

    def test_get_constants(self):
        const = uf.get_constants(self.hist)

        hist_post15140 = uf.compare_version(self.hist.header['version_number'], '15140', '>=')
        const_post15140 = uf.compare_version(const.version, '15140', '>=')

        self.assertTrue(hist_post15140 == const_post15140)  # Both True or both False is correct

    def test_convert_mixing_type(self):
        types = np.arange(10)
        types_pre = uf.convert_mixing_type(types, '11701', -1)
        types_post = uf.convert_mixing_type(types, '15140', -1)

        np.testing.assert_array_equal(np.array([100, 101,  -1, 103, 104, 105, 106,  -1, 107, 109]), types_pre)
        np.testing.assert_array_equal(np.array([100, 101, 103, 104, 105, 106,  -1, 107, 109,  -1]), types_post)

    def test_masks(self):
        for mask_func in uf.mask_functions:
            mask = mask_func(self.hist)
        mask = uf.get_ms_mask(self.hist)
        self.assertEqual(len(self.hist), len(mask))
        self.assertEqual(len(self.hist[mask]), sum(mask))

    def test_calc_deltanu(self):
        gss = ld.load_gss(self.hist)
        gs = gss[0]
        uf.calc_deltanu(gs, self.hist)

    def test_calc_deltaPg(self):
        gss = ld.load_gss(self.hist)
        gs = gss[0]
        uf.calc_deltaPg(gs, self.hist, 1)

    def test_correct_seismo(self):
        uf.correct_seismo(self.hist, ld.load_gss(self.hist, return_pnums=True), uf.get_rgb_mask, 'star_age')




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import matplotlib as mpl
import numpy as np
from matplotlib import cm
from matplotlib import colors
from matplotlib import patheffects
from matplotlib import pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib.legend_handler import HandlerLineCollection
from matplotlib.lines import Line2D
from scipy.interpolate import interp1d

from .. import functions as uf
from ..constants import post15140


class HandlerDashedLines(HandlerLineCollection):
    """
    Custom Handler for LineCollection instances.
    """

    # From https://matplotlib.org/stable/gallery/text_labels_and_annotations/legend_demo.html
    def create_artists(self, legend, orig_handle,
                       xdescent, ydescent, width, height, fontsize, trans):
        # figure out how many lines there are
        numlines = len(orig_handle.get_segments())
        xdata, xdata_marker = self.get_xdata(legend, xdescent, ydescent,
                                             width, height, fontsize)
        leglines = []
        # divide the vertical space where the lines will go
        # into equal parts based on the number of lines
        ydata = np.full_like(xdata, height / (numlines + 1))
        # for each line, create the line at the proper location
        # and set the dash pattern
        for i in range(numlines):
            legline = Line2D(xdata, ydata * (numlines - i) - ydescent)
            self.update_prop(legline, orig_handle, legend)
            # set color, dash pattern, and linewidth to that
            # of the lines in linecollection
            try:
                color = orig_handle.get_colors()[i]
            except IndexError:
                color = orig_handle.get_colors()[0]
            try:
                dashes = orig_handle.get_dashes()[i]
            except IndexError:
                dashes = orig_handle.get_dashes()[0]
            try:
                lw = orig_handle.get_linewidths()[i]
            except IndexError:
                lw = orig_handle.get_linewidths()[0]
            if dashes[1] is not None:
                legline.set_dashes(dashes[1])
            else:
                legline.set_linestyle(orig_handle.get_linestyle()[i])
            legline.set_color(color)
            legline.set_transform(trans)
            legline.set_linewidth(lw)
            leglines.append(legline)
        return leglines


class MidpointBoundaryNorm(colors.BoundaryNorm):
    """
    Generate a colormap index based on discrete intervals.

    Unlike :class:`Normalize` or :class:`LogNorm`,
    :class:`MidpointBoundaryNorm` maps values to integers instead of to the
    interval 0-1.

    Mapping to the 0-1 interval could have been done via
    piece-wise linear interpolation, but using integers seems
    simpler, and reduces the number of conversions back and forth
    between integer and floating point.
    """

    def __init__(self, boundaries, ncolors, midpoint, clip=False, away_from_midpoint=True, singular_midpoint=True):
        """
        Parameters
        ----------
        boundaries : array-like
            Monotonically increasing sequence of boundaries.
        ncolors : int
            Number of colors in the colormap to be used.
        midpoint : scalar
            Value to assign the middle color to.
        clip : bool, optional
            If clip is ``True``, out of range values are mapped to 0 if they
            are below ``boundaries[0]`` or mapped to ncolors - 1 if they are
            above ``boundaries[-1]``.

            If clip is ``False``, out of range values are mapped to -1 if
            they are below ``boundaries[0]`` or mapped to ncolors if they are
            above ``boundaries[-1]``. These are then converted to valid indices
            by :meth:`Colormap.__call__`.
        away_from_midpoint : bool, optional
            If away_from_midpoint is ``True``, determine colors away from
            the midpoint value instead of all in one direction.
        singular_midpoint : bool, optional
            If singular_midpoint is ``True``, only values which equal midpoint
            are set to that color. Only used when away_from_midpoint is ``True``.

        Notes
        -----
        *boundaries* defines the edges of bins, and data falling within a bin
        is mapped to the color with the same index.

        If the number of bins doesn't equal *ncolors*, the color is chosen
        by linear interpolation of the bin number onto color numbers.
        """
        super().__init__(boundaries, ncolors, clip)
        self.away_from_midpoint = away_from_midpoint
        self.singular_midpoint = singular_midpoint
        self.clip = clip
        self.vmin = boundaries[0]
        self.vmax = boundaries[-1]
        self.midpoint = midpoint
        self.boundaries = np.asarray(boundaries)

        self.N = len(self.boundaries)
        self.Ncmap = ncolors
        if self.N - 1 == self.Ncmap:
            self._interp = False
        else:
            self._interp = True

    def __call__(self, value, clip=None):
        if clip is None:
            clip = self.clip

        xx, is_scalar = self.process_value(value)
        mask = np.ma.getmaskarray(xx)
        xx = np.atleast_1d(xx.filled(self.vmax + 1))
        if clip:
            np.clip(xx, self.vmin, self.vmax, out=xx)
            max_col = self.Ncmap - 1
        else:
            max_col = self.Ncmap
        imid = np.argmin(np.abs(self.boundaries - self.midpoint))
        iret = np.zeros(xx.shape, dtype=np.int16)
        if self.away_from_midpoint:
            if self.singular_midpoint:
                left = xx < self.midpoint
                iret[left] = np.digitize(xx[left], self.boundaries) - 1
                right = xx > self.midpoint
                iret[right] = np.digitize(xx[right], self.boundaries, True)
                iret[~(left | right)] = imid  # Exactly equal to midpoint
            else:
                for i, b in enumerate(self.boundaries[::-1]):
                    if b <= self.midpoint:
                        iret[xx <= b] = self.N - 1 - i
                for i, b in enumerate(self.boundaries):
                    if b >= self.midpoint:
                        iret[xx >= b] = i
        else:
            for i, b in enumerate(self.boundaries):
                iret[xx >= b] = i
        if self._interp:
            if self.away_from_midpoint & self.singular_midpoint:
                scalefac_l = self.Ncmap / 2 / imid
                scalefac_r = self.Ncmap / 2 / (self.N - imid + 1)
                imid = int(self.Ncmap / 2)
                iret[left] = (iret[left] * scalefac_l).astype(np.int16)
                iret[right] = (imid + iret[right] * scalefac_r).astype(np.int16)
                iret[~(left | right)] = imid
            else:
                scalefac = (self.Ncmap - 1) / (self.N - 2)
                iret = (iret * scalefac).astype(np.int16)
                imid = int(imid * scalefac)

        if not (self._interp & self.away_from_midpoint & self.singular_midpoint):
            imask = xx < self.midpoint
            iret[imask] = iret[imask] / imid * 0.5 * (self.Ncmap - 1)
            imask = xx >= self.midpoint
            iret[imask] = (iret[imask] - imid) / (self.Ncmap - imid) * 0.5 * (self.Ncmap - 1) + 0.5 * (self.Ncmap - 1)
            iret = iret.astype(np.int16)

        iret[xx < self.vmin] = -1
        iret[xx >= self.vmax] = max_col
        ret = np.ma.array(iret, mask=mask)
        if is_scalar:
            ret = int(ret[0])  # assume python scalar
        return ret

    def inverse(self, value):
        """
        Raises
        ------
        ValueError
            MidpointBoundaryNorm is not invertible, so calling this method will always
            raise an error
        """
        return ValueError("MidpointBoundaryNorm is not invertible")


def get_figure(ax):
    if ax is None:
        f, ax = plt.subplots()
    else:
        f = ax.get_figure()
    return f, ax


def top_legend(ax, ncol=2, **kwargs):
    return ax.legend(ncol=ncol, bbox_to_anchor=(0., 1.02, 1., .102), loc=3, mode='expand', borderaxespad=0., **kwargs)


def top_figure_legend(f, ncol, top=0.9, **kwargs):
    if f.get_constrained_layout():
        return f.legend(loc='outside upper center', ncol=ncol, borderaxespad=0.2, **kwargs)

    fig_size = f.bbox.corners()[3]
    f.subplots_adjust(top=top)
    corners = np.array([ax.bbox.corners() for ax in f.axes]) / fig_size

    left = np.min(corners[:, :, 0])
    right = np.max(corners[:, :, 0])
    bot = np.max(corners[:, :, 1])

    width = right - left
    height = 1 - bot - 0.02

    legend = f.legend(loc='upper center', ncol=ncol, mode='expand', borderaxespad=0.,
                      bbox_to_anchor=(left, bot, width, height),
                      **kwargs)
    return legend


def side_figure_legend(f, ncol, right=0.75, **kwargs):
    fig_size = f.bbox.corners()[3]
    f.subplots_adjust(right=right)
    corners = np.array([ax.bbox.corners() for ax in f.axes]) / fig_size

    left = np.max(corners[:, :, 0]) + 0.01
    top = bot = np.max(corners[:, :, 1])
    bot = np.min(corners[:, :, 1])

    width = 1 - left - 0.01
    height = top - bot

    legend = f.legend(loc='upper center', ncol=ncol, mode='expand', borderaxespad=0.,
                      bbox_to_anchor=(left, bot, width, height),
                      **kwargs)
    return legend


def colored_line(f, ax, xdat, ydat, cdat, norm=None, cmap='viridis', lw=2, add_cbar=True, do_lims=False, **kwargs):
    points = np.asarray([xdat, ydat]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    lc = LineCollection(segments, cmap=plt.get_cmap(cmap), zorder=5, lw=1, norm=norm, **kwargs)
    lc.set_array(cdat)
    lc.set_linewidth(lw)
    ax.add_collection(lc)
    if add_cbar:
        f.colorbar(lc, ax=ax)

    if do_lims:
        # Use automatic x and ylimits
        xmin = min(xdat)
        xmax = max(xdat)
        ymin = min(ydat)
        ymax = max(ydat)

        ax.plot([xmin, ymin], [xmax, ymax], lw=0)
        ax.set_prop_cycle(None)
    return lc


def hrd_const_rad(ax, fontsize=8, radii=None, angle=None, linear=False):
    xlim1 = ax.get_xlim()
    ylim1 = ax.get_ylim()
    if linear:
        xlim1 = np.log10(xlim1)
        ylim1 = np.log10(ylim1)
    xw = xlim1[0] - xlim1[1]
    yw = ylim1[1] - ylim1[0]

    xfactor = 0.07 / 4 * yw / xw

    # at this scale doesn't matter which constants are used
    const = 4 * np.pi * post15140.boltz_sigma / post15140.lsun * post15140.rsun ** 2
    log_const = np.log10(const)
    logTs = np.linspace(0, 6, 2)
    if radii is None:
        radii = np.logspace(-1, 6, 8)

    radii = np.unique(radii)
    radii = radii[radii > 0]
    _radii = []
    for rad in radii:
        # Do number string formatting first so calculated and displayed radii are equal
        if max(radii) < 10:
            nr_string = f'{rad:.1f}'
        else:
            if rad < 1:
                nr_string = f'{rad:.1f}'
            else:
                nr_string = f'{rad:.0f}'
        rad = float(nr_string)
        if (rad in _radii) or (rad <= 0):
            continue
        _radii.append(rad)

        logR = np.log10(rad)
        logLs = 2 * logR + 4 * logTs + log_const
        if linear:
            ax.plot(10**logTs, 10**logLs, 'grey', linestyle='--', zorder=-10, lw=1)
        else:
            ax.plot(logTs, logLs, 'grey', linestyle='--', zorder=-10, lw=1)

        xleft = np.interp(ylim1[1], logLs, logTs)
        xright = np.interp(ylim1[0], logLs, logTs)
        xleft = min(xleft, xlim1[0])
        xright = max(xright, xlim1[1])

        yleft = np.interp(xleft, logTs, logLs)
        yright = np.interp(xright, logTs, logLs)

        is_top = (xleft < xlim1[0])
        is_bottom = (xright > xlim1[1])
        textxs = []

        if is_top:
            textxs.append(xleft - xfactor * xw)
            textxs.append(xright + 0.07 * xw)
        elif is_bottom:
            textxs.append(xright + xfactor * xw)
            textxs.append(xleft - 0.07 * xw)
        else:
            textxs.append(xleft - 0.07 * xw)
            textxs.append(xright + 0.07 * xw)
        frac_x_sep = abs(np.diff(textxs)) / xw
        if frac_x_sep < 0.1:
            textxs.pop(0)

        for i, textx in enumerate(textxs):

            if angle is None:
                # Have to reset xylims due to the plotting the lines of constant radius, otherwise get wrong angles.
                if linear:
                    ax.set_ylim(10**ylim1)
                    ax.set_xlim(10**xlim1)
                    screen_dx, screen_dy = ax.transData.transform((10**logTs[0], 10**logLs[0])) - ax.transData.transform(
                        (10**logTs[-1], 10**logLs[-1]))
                else:
                    ax.set_ylim(ylim1)
                    ax.set_xlim(xlim1)
                    screen_dx, screen_dy = ax.transData.transform((logTs[0], logLs[0])) - ax.transData.transform(
                        (logTs[-1], logLs[-1]))
                angle = (np.degrees(np.arctan2(screen_dy, screen_dx)) + 90) % 180 - 90
            texty = np.interp(textx, logTs, logLs)
            if linear:
                textx = 10**textx
                texty = 10**texty
            text = ax.text(textx, texty, r"${}\;\mathrm{{R}}_{{\odot}}$".format(nr_string), color='k', zorder=-9,
                           rotation=angle, size=fontsize, va='center', ha='center', clip_on=True,
                           bbox=dict(facecolor='white', linewidth=0, pad=0))
            text.set_path_effects(
                [patheffects.Stroke(linewidth=3, foreground=ax.get_facecolor()), patheffects.Normal()])
    if linear:
        ax.set_ylim(10**ylim1)
        ax.set_xlim(10**xlim1)
    else:
        ax.set_ylim(ylim1)
        ax.set_xlim(xlim1)


def get_x_and_set_xlabel(p, xname, ax=None, func_on_xaxis=None, hist=None):
    if isinstance(xname, list) or isinstance(xname, tuple):
        x, xname = xname
        label = xname
    elif xname.lower().replace('_', '') in ['radius', 'r', 'rsol', 'rsun']:
        x = uf.get_radius(p, 'Rsol')
        label = r'Radius $(R_\odot)$'
    elif xname in ['radius_cm', 'r_cm']:
        x = uf.get_radius(p)
        label = r'Radius (cm)'
    elif xname in ('x', 'radius_dimless'):
        x = uf.get_radius(p)
        x = x / max(x)
        label = r'Fractional radius $x$ $( r /\mathrm{R}_\star)$'
    elif xname == 'logR':
        x = uf.get_radius(p, 'Rsol')
        label = r'Radius $(r/\mathrm{R}_\odot)$'
    elif xname == 'mass':
        x = p.data.mass
        label = r'Mass coordinate $m/\mathrm{M}_\odot$'
    elif xname in ['q', 'mass_dimless']:
        x = p.data.mass
        x = x / x[0]
        label = r'Fractional mass $q$ $( m /\mathrm{M}_\star)$'
    elif xname in ['zone', 'zone_number']:
        x = np.arange(len(p.data)) + 1
        label = 'Zone Number'
    elif xname == 's':
        if hist is None:
            raise ValueError("Need History for xname='s'.")
        i_hist = p.get_hist_index(hist)
        r1 = hist.data.r_1[i_hist]
        r2 = hist.data.r_2[i_hist]
        r0 = np.sqrt(r1 * r2)
        x = np.log(p.data.radius / r0)
        label = '$s$'
    else:
        x = p.get(xname)
        label = xname.replace('_', ' ')
    if ax is not None:
        ax.set_xlabel(label)

    if func_on_xaxis is not None:
        x = func_on_xaxis(x)
    return x


def get_mix_dict(profile):
    mesa_ver = str(profile.header.get('version_number'))
    if uf.compare_version(mesa_ver, '15140', '>='):
        prefix = 'post'
    else:
        prefix = 'pre'
    mix_dict = uf.mix_dict[f'{prefix}15140']
    return mix_dict


def get_mixing(profile, min_width):
    """

    :param profile:
    :param min_width:
    :return:
    """
    mix_dict = get_mix_dict(profile)

    mixing = profile.get('mixing_type')
    regions = {}
    for i, name in mix_dict.items():
        # if i == 0:
        #    continue
        good_zones = np.where(mixing == i)[0]
        changes = np.where(np.diff(good_zones) != 1)[0]

        starts = []
        ends = []
        if len(good_zones) > 1:
            starts.append(good_zones[0])
            for change in changes:
                ends.append(good_zones[change] + 1)  # Add one to include final cell
                starts.append(good_zones[change + 1])
            ends.append(good_zones[-1])
        pairs = zip(starts, ends)
        regions[name] = [pair for pair in pairs if pair[1] - pair[0] >= min_width]
    return regions

def get_default_mixing_kwargs():
    """
    Get the default coloring scheme for mixing.
    Returns:
        dict of dict: Each dict contains kwargs for PathPatch and `show`, which if `False` mean that it will not be
            drawn.
    """
    default_kwargs_mixing = {
        uf.mix_dict['merged']['convective_mixing']: {'name': 'convective_mixing',
                                                     'color': "Chartreuse",
                                                     'hatch': "//",
                                                     'line': 1,
                                                     'show': True
                                                     },
        uf.mix_dict['merged']['overshoot_mixing']: {'name': 'overshoot_mixing',
                                                    'color': "purple",
                                                    'hatch': "x",
                                                    'line': 1,
                                                    'show': True
                                                    },
        uf.mix_dict['merged']['semiconvective_mixing']: {'name': 'semiconvective_mixing',
                                                         'color': "red",
                                                         'hatch': "\\\\",
                                                         'line': 1,
                                                         'show': True
                                                         },
        uf.mix_dict['merged']['thermohaline_mixing']: {'name': 'thermohaline_mixing',
                                                       'color': "Gold",
                                                       'hatch': "||",
                                                       'line': 1,
                                                       'show': False
                                                       },
        uf.mix_dict['merged']['rotation_mixing']: {'name': 'rotation_mixing',
                                                   'color': "brown",
                                                   'hatch': "*",
                                                   'line': 1,
                                                   'show': True
                                                   },
        uf.mix_dict['merged']['anonymous_mixing']: {'name': 'anonymous_mixing',
                                                    'color': "tab:grey",
                                                    'hatch': "+",
                                                    'line': 1,
                                                    'show': True
                                                    },
        uf.mix_dict['merged']['minimum_mixing']: {'name': 'minimum_mixing',
                                                  'color': "cyan",
                                                  'hatch': "-",
                                                  'line': 1,
                                                  'show': True
                                                  },
        uf.mix_dict['merged']['no_mixing']: {'name': 'no_mixing',
                                             'color': "",
                                             'hatch': "",
                                             'line': 0,
                                             'show': False
                                             },

    }
    return default_kwargs_mixing

def add_mixing(ax, profile, xname='mass', min_width=5, ymin=0, ymax=1, alpha=1, add_legend=True, func_on_xaxis=None,
               kwargs_mixing=None):
    """
    kwargs_mixing (dict, optional): kwargs used to draw mixing regions, if `None`, defaults to `plotting.utils.get_default_mixing_kwargs()`.
    """
    if kwargs_mixing is None:
        kwargs_mixing = get_default_mixing_kwargs()

    x = get_x_and_set_xlabel(profile, xname, func_on_xaxis=func_on_xaxis)

    mix_dict = get_mix_dict(profile)
    all_regions = get_mixing(profile, min_width)
    added_to_legend = []
    for _, name in mix_dict.items():
        regions = all_regions[name]
        if len(regions) == 0:
            continue
        mix_type = uf.mix_dict['merged'][name]
        mix_info = kwargs_mixing[mix_type]
        color = mix_info['color']
        hatch = mix_info['hatch']
        line = mix_info['line']
        show = mix_info['show']

        if not show:
            continue

        for region in regions:
            start, end = region
            xstart = x[start]
            xend = x[end]

            if add_legend and name not in added_to_legend:
                ax.axvspan(xstart, xend, ymin, ymax, hatch=hatch, edgecolor=color, linewidth=line,
                           facecolor='none', label=name, zorder=-1, alpha=alpha)
                added_to_legend.append(name)
            else:
                ax.axvspan(xstart, xend, ymin, ymax, hatch=hatch, edgecolor=color, linewidth=line,
                           facecolor='none', zorder=-1, alpha=alpha)


def add_burning(ax, profile, xname='mass', ymin=0, ymax=1, num_levels=None, vmin=-2, vmax=8, add_cbar=True,
                kind='net_nuclear_energy', norm=None, kipp_scaling=True, func_on_xaxis=None):
    """
    """
    x = get_x_and_set_xlabel(profile, xname, func_on_xaxis=func_on_xaxis)

    if kind == 'eps_nuc':
        log_z = np.log10(profile.get('eps_nuc'))
        cbar_label = r'$\log_{10} \; \epsilon_{\mathrm{nuc}}$'
    elif kind == 'net_nuclear_energy':
        log_z = profile.get('net_nuclear_energy')
        if kipp_scaling:
            log_z = np.sign(log_z) * np.ceil(np.abs(log_z))
        cbar_label = r'$\mathrm{sign}(\epsilon_\mathrm{net}) \; \log_{10} \; \max(1, \lceil|\epsilon_\mathrm{net}|\rceil)/(\mathrm{erg}/g/s)$'
    else:
        raise ValueError('`eps_nuc` or `net_nuclear_energy` must be a column in profile.')
    # log_z[log_z < vmin] = vmin
    if num_levels is None:
        num_levels = int(vmax - vmin + 1)

    if (kind == 'net_nuclear_energy') and (vmin < 0):
        levels = np.linspace(vmin - 0.5, vmax + 0.5, num_levels + 1)
        cmap = plt.cm.RdBu
        if norm is None:
            norm = MidpointBoundaryNorm(levels, cmap.N, midpoint=0, singular_midpoint=True)
    else:
        levels = np.linspace(vmin, vmax, num_levels)
        cmap = plt.cm.Blues
        if norm is None:
            norm = colors.BoundaryNorm(levels, cmap.N)

    log_z_disc = norm(log_z)
    i_max = len(log_z_disc) - 1
    pairs = []
    for level in levels:
        mask = log_z_disc == norm(level)
        diff = np.diff(mask.astype(int))
        transitions = list(np.where(diff != 0)[0])
        if len(transitions) == 0:
            continue

        if mask[0]:
            transitions = [0] + transitions
        if mask[-1]:
            transitions = transitions + [i_max]
        for i in range(len(transitions) // 2):
            i_left = transitions[2 * i]
            i_right = transitions[2 * i + 1]
            pairs.append((level, (i_left, i_right)))

    for level, (i_left, i_right) in pairs:
        ax.axvspan(x[i_left], x[i_right], ymin, ymax, color=cmap(norm(level)), ec=None, zorder=-2)

    if add_cbar:
        f = ax.get_figure()
        f.colorbar(cm.ScalarMappable(cmap=cmap, norm=norm), ticks=np.linspace(vmin, vmax, num_levels)[1::2], ax=ax)
        cbax = f.axes[-1]
        cbax.set_ylabel(cbar_label)

    return cm.ScalarMappable(cmap=cmap, norm=norm), cbar_label


def add_hrd_instabilities(ax, classic=True, sdB=True, ):
    xlims = ax.get_xlim()
    ylims = ax.get_ylim()

    if classic:
        ax.plot((3.95, 3.9, 3.8), (1, 1.8, 4.5), 'k--')
        ax.plot((3.85, 3.8, 3.7), (0.8, 1.6, 4.5), 'k--')
    if sdB:
        sdBVr_logTeff = (4.55, 4.45)
        sdBVr_logL = (1, 1.8)
        patch_sdBVr = mpl.patches.Rectangle((sdBVr_logTeff[0], sdBVr_logL[0]),
                                            width=np.diff(sdBVr_logTeff), height=np.diff(sdBVr_logL),
                                            fill=False, edgecolor='b', hatch='\\')
        ax.add_patch(patch_sdBVr)

        sdBVs_logTeff = (4.5, 4.35)
        sdBVs_logL = (1.4, 1.8)
        patch_sdBVr = mpl.patches.Rectangle((sdBVs_logTeff[0], sdBVs_logL[0]),
                                            width=np.diff(sdBVs_logTeff), height=np.diff(sdBVs_logL),
                                            fill=False, edgecolor='b', hatch='/')
        ax.add_patch(patch_sdBVr)

    ax.set_xlim(xlims)
    ax.set_ylim(ylims)


def calc_inertia_marker_size(gs, l, freq_units='uHz'):
    mask = gs.get('l') == 0
    E_l0 = gs.get('E_norm')[mask]
    nu_all = gs.get_frequencies(freq_units)
    nu_l0 = nu_all[mask]
    # interpolate over log10 inertia for better behaviour
    try:
        log_f_El0 = interp1d(nu_l0, np.log10(E_l0), kind='cubic', bounds_error=True)
    except ValueError:
        log_f_El0 = interp1d(nu_l0, np.log10(E_l0), kind='linear', bounds_error=True)
    mask = gs.get('l') == l
    nu = gs.get_frequencies(freq_units)[mask]
    # x = np.log10(gs.get('E_norm')[mask]) - log_f_El0(nu_all[mask])
    # ms = 2.5 * 10 ** (2 * (1 - x))
    xmin = min(E_l0)
    x = np.log10(gs.get('E_norm')[mask] / xmin)
    ms = 25 - 2 * x ** 3
    ms = np.minimum(25, ms)
    ms = np.maximum(1, ms)
    return ms


def line_legend(ax, edge_space=0.05, num_line_label=4, fontsize=7, background=None, background_width=3, ignore_nolabel=False):
    xmin, xmax = ax.get_xlim()
    x_range = xmax - xmin
    line_label_x = np.linspace(xmin + edge_space * x_range, xmax - edge_space * x_range, num=num_line_label)
    if background is None:
        background = ax.get_facecolor()
    # Partially based on https://github.com/cphyc/matplotlib-label-lines
    for line in ax.get_lines():
        label = line.get_label()
        xdata = line.get_xdata()
        ydata = line.get_ydata()

        if ignore_nolabel:
            if label.startswith('_child'):
                continue

        for x in line_label_x:
            idxs = np.where(np.diff(np.sign(xdata - x)) != 0)[0] + 1
            for idx in idxs:
                # Data ordered from surface to core
                x_left = xdata[idx]
                x_right = xdata[idx - 1]
                y_left = ydata[idx]
                y_right = ydata[idx - 1]
                y = np.interp(x, (x_left, x_right), (y_left, y_right))

                screen_dx, screen_dy = ax.transData.transform((x_left, y_left)) - ax.transData.transform(
                    (x_right, y_right))
                rotation = (np.degrees(np.arctan2(screen_dy, screen_dx)) + 90) % 180 - 90

                text = ax.text(x, y, label, rotation=rotation, color=line.get_color(), ha='center', va='center',
                               clip_on=True, size=fontsize, bbox={'alpha': 0})
                text.set_path_effects(
                    [patheffects.Stroke(linewidth=background_width, foreground=background),
                     patheffects.Normal()])



def decimate_RDP(pts, epsilon, return_index=False):
    """
    Ramer–Douglas–Peucker decimation algorithm with some adjustments.
    Args:
        pts (np.array): 2-d array containing points to decimate.
        epsilon: Tolerance for keeping line segments
        return_index (bool, optional): If True, return indeces instead of values. Defaults to False.

    Returns:
        np.array: Decimated version of pts.
    """
    good = np.isfinite(pts[:,1]) & (pts[:,1] >= 0) & (pts[:,1] <= 1)
    good_loc = np.where(good)[0]
    last_i = good_loc[-1]
    # First and last non-nan point
    i_start = good_loc[0]
    i_end = good_loc[-1]

    new_pts = np.zeros(len(pts), dtype=int)
    new_pts[0] = i_start
    i_insert = 1

    while i_start <= last_i:
        pt0, pt1 = pts[i_start], pts[i_end]
        if np.isnan(pt0 + pt1).any():  # Keep one NaN as separator for blocks
            new_pts[i_insert] = i_end
            i_insert += 1
            i_start = good_loc[good_loc > i_end][0]  # First good point after current end
            i_end = last_i
            pt0, pt1 = pts[i_start], pts[i_end]

        if i_end - i_start <= 1:  # If next point has difference larger than epsilon keep it and start from next.
            new_pts[i_insert] = i_end
            i_insert += 1
            i_start += 1
            i_end = last_i
            pt0, pt1 = pts[i_start], pts[i_end]

        # Perpendicular distance
        delta = np.abs(np.cross(pt1 - pt0, pt0 - pts[i_start + 1:i_end]) / np.linalg.norm(pt1 - pt0))
        try:
            i_dmax = np.nanargmax(delta)
            dmax = delta[i_dmax]
        except ValueError:
            if np.all(np.isnan(delta)):
                dmax = 0
            else:
                print('Bad delta in RDP decimation')
                print('i_start, i_end, pt0, pt1')
                print(i_start, i_end, pt0, pt1)
                print('np.linalg.norm(pt1 - pt0)')
                print(np.linalg.norm(pt1 - pt0))
                print('delta')
                print(delta)
                raise

        if dmax <= epsilon:  # Keep i_end and skip points in between
            new_pts[i_insert] = i_end
            i_insert += 1
            i_start = i_end
            i_end = last_i
        else:
            i_end = i_start + i_dmax + 1
        if i_start >= last_i:  # Done
            break
    new_pts = new_pts[:i_insert]
    new_pts = np.sort(new_pts)
    new_pts = new_pts[(new_pts >= 0) & (new_pts < len(pts))]
    if return_index:
        return new_pts
    else:
        return pts[new_pts]



#/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import profile
import unittest
import numpy as np

from wsssss import load_data as ld
from .common import have_mesa_data

have_mesa_data()
test_data = os.path.join(os.path.dirname(__file__), 'data', 'mesa')
curdir = os.path.dirname(__file__)
class TestLoadData(unittest.TestCase):

    def test_History(self):
        hist = ld.History(os.path.join(test_data, '0000', 'LOGS', 'history.data'))
        hist.dump()

        hist_dill = ld.History(os.path.join(test_data, '0000', 'LOGS', 'history.data.dill'))
        np.testing.assert_array_equal(hist.data, hist_dill.data)
        self.assertDictEqual(hist.header, hist_dill.header)
        del hist_dill

        hist_reload = ld.History(os.path.join(test_data, '0000', 'LOGS', 'history.data'), save_dill=True, reload=True)
        np.testing.assert_array_equal(hist.data, hist_reload.data)
        self.assertDictEqual(hist.header, hist_reload.header)
        del hist_reload

        np.testing.assert_array_equal(np.arange(1, 1001), hist.get('model_number'))
        np.testing.assert_array_equal(np.arange(1, 1001), hist.data.model_number)
        np.testing.assert_array_equal(np.ones(1000), hist.data.star_mass)
        np.testing.assert_array_equal(np.arange(1, 11), hist[:10].data.model_number)

        np.testing.assert_array_equal(hist.index[:, 0]-1, hist.get_profile_index(hist.index[:, 2]))
        np.testing.assert_array_equal((2, 100, 99), hist.get_profile_num(150))
        np.testing.assert_array_equal((2, 100, 99), hist.get_profile_num(150, method='previous'))
        np.testing.assert_array_equal((3, 200, 199), hist.get_profile_num(150, method='next'))
        np.testing.assert_array_equal((3, 200, 199), hist.get_profile_num(150, earlier=False))

        hist_cols = ld.History(os.path.join(test_data, '0000', 'LOGS', 'history.data'), keep_columns=['model_number', 'center_he4'])
        self.assertListEqual(['model_number', 'center_he4'], hist_cols.columns)
        self.assertListEqual(hist_cols.columns, list(hist_cols.data.dtype.names))
        np.testing.assert_array_equal(hist.data[hist_cols.columns], hist_cols.data[hist_cols.columns])

    def test_Profile(self):
        prof = ld.Profile(os.path.join(test_data, '0000', 'LOGS', 'profile1.data'))
        hist = ld.History(os.path.join(test_data, '0000', 'LOGS', 'history.data'))

        np.testing.assert_array_equal(np.zeros(1), hist.get_profile_index(prof))
        np.testing.assert_array_equal(np.zeros(1), hist.get_profile_index([prof]))
        np.testing.assert_array_equal(np.zeros(1), hist.get_profile_index(prof.profile_num))
        np.testing.assert_array_equal(np.zeros(1), hist.get_profile_index([prof.profile_num]))

        prof = ld.Profile(os.path.join(test_data, '0000', 'LOGS', 'profile1.data'), load_GyreProfile=True)

        self.assertEqual(prof.get_hist_index(hist), 0)

    def test_GyreSummary(self):
        gsum = ld.GyreSummary(os.path.join(test_data, '0000', 'gyre_out', 'profile10.data.GYRE.sgyre_l'))
        self.assertEqual(7, len(gsum.data[gsum.data['l'] == 0]))
        self.assertEqual(236, len(gsum.data[gsum.data['l'] == 1]))

    def test_GyreProfile(self):
        prof = ld.Profile(os.path.join(test_data, '0000', 'LOGS', 'profile1.data'))
        gprof = ld.GyreProfile(os.path.join(test_data, '0000', 'LOGS', 'profile1.data.GYRE'))

        np.testing.assert_allclose(prof.data.mass/prof.data.mass[0], np.interp(prof.data.radius/prof.data.radius[0], gprof.data.radius/gprof.header['star_radius'], gprof.data.mass/gprof.header['star_mass']), rtol=1e-11)

    def test_GyreMode(self):
        gsum = ld.GyreSummary(os.path.join(test_data, '0000', 'gyre_out', 'profile10.data.GYRE.sgyre_l'))
        gmode = ld.GyreSummary(os.path.join(test_data, '0000', 'gyre_out', 'profile10.data.GYRE_l0_00005_np+9_ng+0.mgyre'))
        np = 9
        ng = 0
        mask = (gsum.data.n_p == np) & (gsum.data.n_g == ng)
        self.assertEqual(1, sum(mask))
        self.assertEqual(gmode.header['Re(freq)'], gsum.data['Re(freq)'][mask])

    def test_load_profs(self):
        hist = ld.History(os.path.join(test_data, '0000', 'LOGS', 'history.data'))
        profs = ld.load_profs(hist)
        self.assertEqual(11, len(profs))
        self.assertListEqual(list(np.arange(1, 12)), [prof.profile_num for prof in profs])

    def test_load_gss(self):
        hist = hist = ld.History(os.path.join(test_data, '0000', 'LOGS', 'history.data'))
        gss = ld.load_gss(hist)
        self.assertEqual(11, len(gss))

    def load_modes_from_profile(self):
        gsum = ld.GyreSummary(os.path.join(test_data, '0000', 'gyre_out', 'profile10.data.GYRE.sgyre_l'))
        prof = ld.Profile(os.path.join(test_data, '0000', 'LOGS', 'profile10.data'))
        modes = ld.load_modes_from_profile(prof)
        self.assertEqual(7, len(gsum.data[gsum.data.l==0]))

    def load_gs_from_profile(self):
        prof = ld.Profile(os.path.join(test_data, '0000', 'LOGS', 'profile10.data'))
        gs = ld.load_gs_from_profile(prof)
        self.assertEqual(243, len(gs.data))

